# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from os.path import join
from uuid import uuid4
try:
    from itertools import izip_longest
except ImportError:
    from itertools import zip_longest as izip_longest

from django.conf import settings
from django.utils.module_loading import import_string

from eoxserver.core.util.iteratortools import pairwise_iterative
from eoxserver.contrib import mapserver as ms
from eoxserver.contrib import vsi, vrt, gdal, osr
from eoxserver.render.browse.objects import (
    Browse, GeneratedBrowse, BROWSE_MODE_GRAYSCALE
)
from eoxserver.render.browse.generate import (
    generate_browse, FilenameGenerator
)
from eoxserver.render.map.objects import (
    CoverageLayer, CoveragesLayer, MosaicLayer, OutlinedCoveragesLayer,
    BrowseLayer, OutlinedBrowseLayer,
    MaskLayer, MaskedBrowseLayer, OutlinesLayer
)
from eoxserver.render.mapserver.config import (
    DEFAULT_EOXS_MAPSERVER_LAYER_FACTORIES,
)
from eoxserver.render.colors import BASE_COLORS, COLOR_SCALES, OFFSITE_COLORS
from eoxserver.resources.coverages import crss
from eoxserver.resources.coverages.dateline import (
    extent_crosses_dateline, wrap_extent_around_dateline
)
from eoxserver.processing.gdal import reftools

import logging
logger = logging.getLogger(__name__)


class BaseMapServerLayerFactory(object):
    handled_layer_types = []

    @classmethod
    def supports(self, layer_type):
        return layer_type in self.handled_layer_types

    def create(self, map_obj, layer):
        pass

    def destroy(self, map_obj, layer, data):
        pass


class CoverageLayerFactoryMixIn(object):
    """ Base class for factories dealing with coverages.
    """
    def get_fields(self, fields, bands, wavelengths):
        """ Get the field subset for the given bands/wavelengths selection
        """
        if bands:
            assert len(bands) in (1, 3, 4)
            try:
                fields = [
                    next(field for field in fields if field.identifier == band)
                    for band in bands
                ]
            except StopIteration:
                raise Exception('Invalid bands specified.')
        elif wavelengths:
            assert len(wavelengths) in (1, 3, 4)
            try:
                fields = [
                    next(
                        field
                        for field in fields if field.wavelength == wavelength
                    )
                    for wavelength in wavelengths
                ]
            except StopIteration:
                raise Exception('Invalid wavelengths specified.')
        else:
            # when fields is not 1 (single band grayscale), 3 (RGB) or 4 (RGBA)
            # then use the first band by default
            if len(fields) not in (1, 3, 4):
                return fields[:1]

        return fields

    def create_coverage_layer(self, map_obj, coverage, fields,
                              style=None, ranges=None):
        """ Creates a mapserver layer object for the given coverage
        """
        filename_generator = FilenameGenerator(
            '/vsimem/{uuid}.{extension}', 'vrt'
        )

        field_locations = [
            (field, coverage.get_location_for_field(field))
            for field in fields
        ]
        locations = [
            location
            for _, location in field_locations
        ]

        # TODO: apply subsets in time/elevation dims
        num_locations = len(set(locations))
        if num_locations == 1:
            if not coverage.grid.is_referenceable:
                location = field_locations[0][1]
                data = location.path
                ms.set_env(map_obj, location.env, True)
            else:
                vrt_path = filename_generator.generate()
                e = map_obj.extent
                resx = (e.maxx - e.minx) / map_obj.width
                resy = (e.maxy - e.miny) / map_obj.height

                wkt = osr.SpatialReference(map_obj.getProjection()).wkt

                # TODO: env?
                reftools.create_rectified_vrt(
                    field_locations[0][1].path, vrt_path,
                    order=1, max_error=10,
                    resolution=(resx, -resy), srid_or_wkt=wkt
                )
                data = vrt_path

        elif num_locations > 1:
            paths_set = set(
                field_location[1].path for field_location in field_locations
            )
            if len(paths_set) == 1:
                location = field_locations[0][1]
                data = location.path
                ms.set_env(map_obj, location.env, True)

            else:
                # TODO
                _build_vrt(coverage.size, field_locations)

        if not coverage.grid.is_referenceable:
            extent = coverage.extent
            sr = coverage.grid.spatial_reference
        else:
            map_extent = map_obj.extent
            extent = (
                map_extent.minx, map_extent.miny,
                map_extent.maxx, map_extent.maxy
            )
            sr = osr.SpatialReference(map_obj.getProjection())

        layer_objs = _create_raster_layer_objs(
            map_obj, extent, sr, data, filename_generator
        )

        for i, layer_obj in enumerate(layer_objs):
            layer_obj.name = '%s__%d' % (coverage.identifier, i)
            layer_obj.setProcessingKey("CLOSE_CONNECTION", "CLOSE")

        if num_locations == 1:
            for layer_obj in layer_objs:
                layer_obj.setProcessingKey("BANDS", ",".join([
                    str(coverage.get_band_index_for_field(field))
                    for field in fields
                ]))
        elif num_locations > 1:
            for layer_obj in layer_objs:
                if len(field_locations) == 3:
                    layer_obj.setProcessingKey("BANDS", "1,2,3")
                else:
                    layer_obj.setProcessingKey("BANDS", "1")

        # make a color-scaled layer
        if len(fields) == 1:
            field = fields[0]
            if ranges:
                range_ = ranges[0]
            else:
                range_ = _get_range(field)

            for layer_obj in layer_objs:
                _create_raster_style(
                    style or "blackwhite", layer_obj, range_[0], range_[1], [
                        nil_value[0] for nil_value in field.nil_values
                    ]
                )
        elif len(fields) in (3, 4):
            for i, field in enumerate(fields, start=1):
                if ranges:
                    if len(ranges) == 1:
                        range_ = ranges[0]
                    else:
                        range_ = ranges[i - 1]
                else:
                    range_ = _get_range(field)

                for layer_obj in layer_objs:
                    layer_obj.setProcessingKey(
                        "SCALE_%d" % i,
                        "%s,%s" % range_
                    )
                    layer_obj.offsite = ms.colorObj(0, 0, 0)

        else:
            raise Exception("Too many bands specified")

        return filename_generator

    def destroy_coverage_layer(self, filename_generator):
        for filename in filename_generator.filenames:
            vsi.unlink(filename)


class CoverageLayerFactory(CoverageLayerFactoryMixIn,
                           BaseMapServerLayerFactory):
    handled_layer_types = [CoverageLayer, CoveragesLayer]

    def create(self, map_obj, layer):
        if isinstance(layer, CoverageLayer):
            coverages = [layer.coverage]
        else:
            coverages = layer.coverages

        coverage_layers = []
        for coverage in reversed(coverages):
            fields = self.get_fields(
                coverage.range_type, layer.bands, layer.wavelengths
            )
            coverage_layers.append(
                self.create_coverage_layer(
                    map_obj, coverage, fields, layer.style, layer.ranges
                )
            )

        return coverage_layers

    def destroy(self, map_obj, layer, data):
        for coverage_layer in data:
            self.destroy_coverage_layer(coverage_layer)


class OutlinedCoverageLayerFactory(CoverageLayerFactoryMixIn,
                                   BaseMapServerLayerFactory):
    handled_layer_types = [OutlinedCoveragesLayer]

    def create(self, map_obj, layer):
        coverages = layer.coverages
        style = layer.style

        raster_style = (
            style if style and style in COLOR_SCALES else "blackwhite"
        )
        vector_style = (
            style if style and style in BASE_COLORS else "red"
        )

        coverage_layers = []

        for coverage in reversed(coverages):
            fields = self.get_fields(
                coverage.range_type, layer.bands, layer.wavelengths
            )
            coverage_layers.append(
                self.create_coverage_layer(
                    map_obj, coverage, fields, raster_style, layer.ranges
                )
            )

            # create the outlines layer
            outlines_layer_obj = _create_polygon_layer(map_obj)
            shape_obj = ms.shapeObj.fromWKT(coverage.footprint.wkt)
            outlines_layer_obj.addFeature(shape_obj)

            class_obj = _create_geometry_class(vector_style)
            outlines_layer_obj.insertClass(class_obj)

        return coverage_layers

    def destroy(self, map_obj, layer, data):
        for coverage_layer in data:
            self.destroy_coverage_layer(coverage_layer)


class MosaicLayerFactory(CoverageLayerFactoryMixIn, BaseMapServerLayerFactory):
    handled_layer_types = [MosaicLayer]

    def create(self, map_obj, layer):
        mosaic = layer.mosaic
        fields = self.get_fields(
            mosaic.range_type, layer.bands, layer.wavelengths
        )
        return [
            self.create_coverage_layer(
                map_obj, coverage, fields, layer.style, layer.ranges
            )
            for coverage in reversed(layer.coverages)
        ]

    def destroy(self, map_obj, layer, data):
        for item in data:
            self.destroy_coverage_layer(item)

# TODO: combine BrowseLayerFactory with OutlinedBrowseLayerFactory, as they are
# very similar


class BrowseLayerMixIn(object):
    def make_browse_layer_generator(self, map_obj, browses, map_,
                                    filename_generator, group_name, ranges,
                                    style):
        for browse in browses:
            if isinstance(browse, GeneratedBrowse):
                creation_info, filename_generator, reset_info = \
                    generate_browse(
                        browse.band_expressions,
                        browse.fields_and_coverages,
                        map_.width, map_.height,
                        map_.bbox,
                        map_.crs,
                        filename_generator
                    )
                layer_objs = _create_raster_layer_objs(
                    map_obj, browse.extent, browse.spatial_reference,
                    creation_info.filename, filename_generator
                )

                for layer_obj in layer_objs:
                    layer_obj.data = creation_info.filename
                    if creation_info.env:
                        ms.set_env(map_obj, creation_info.env, True)

                    if creation_info.bands:
                        layer_obj.setProcessingKey('BANDS', ','.join(
                            str(band) for band in creation_info.bands
                        ))

                    if reset_info:
                        sr = osr.SpatialReference(map_.crs)
                        extent = map_.bbox
                        layer_obj.setMetaData(
                            "wms_extent",
                            "%f %f %f %f" % extent
                        )
                        layer_obj.setExtent(*extent)

                        if sr.srid is not None:
                            short_epsg = "EPSG:%d" % sr.srid
                            layer_obj.setMetaData("ows_srs", short_epsg)
                            layer_obj.setMetaData("wms_srs", short_epsg)
                        layer_obj.setProjection(sr.proj)

                if browse.mode == BROWSE_MODE_GRAYSCALE:
                    field = browse.field_list[0]
                    if ranges:
                        browse_range = ranges[0]
                    elif browse.ranges[0] != (None, None):
                        browse_range = browse.ranges[0]
                    else:
                        browse_range = _get_range(field)

                    for layer_obj in layer_objs:
                        _create_raster_style(
                            style or "blackwhite", layer_obj,
                            browse_range[0], browse_range[1], [
                                nil_value[0] for nil_value in field.nil_values
                            ]
                        )

                else:
                    browse_iter = enumerate(
                        zip(browse.field_list, browse.ranges), start=1
                    )
                    for i, (field, field_range) in browse_iter:
                        if ranges:
                            if len(ranges) == 1:
                                range_ = ranges[0]
                            else:
                                range_ = ranges[i - 1]
                        elif field_range != (None, None):
                            range_ = field_range
                        else:
                            range_ = _get_range(field)

                        for layer_obj in layer_objs:
                            layer_obj.setProcessingKey(
                                "SCALE_%d" % i,
                                "%s,%s" % tuple(range_)
                            )

            elif isinstance(browse, Browse):
                layer_objs = _create_raster_layer_objs(
                    map_obj, browse.extent, browse.spatial_reference,
                    browse.filename, filename_generator
                )
                for layer_obj in layer_objs:
                    layer_obj.data = browse.filename
                ms.set_env(map_obj, browse.env, True)
            elif browse is None:
                # TODO: figure out why and deal with it?
                continue
            else:
                raise TypeError(
                    'Type %s is not supported', type(browse).__name__
                )

            for layer_obj in layer_objs:
                layer_obj.group = group_name

            yield browse, layer_objs


class BrowseLayerFactory(CoverageLayerFactoryMixIn, BrowseLayerMixIn,
                         BaseMapServerLayerFactory):
    handled_layer_types = [BrowseLayer]

    def create(self, map_obj, layer):
        filename_generator = FilenameGenerator(
            '/vsimem/{uuid}.{extension}', 'vrt'
        )
        group_name = layer.name
        ranges = layer.ranges
        style = layer.style

        generator = self.make_browse_layer_generator(
            map_obj, reversed(layer.browses), layer.map, filename_generator,
            group_name, ranges, style
        )

        for _ in generator:
            pass

        return filename_generator

    def destroy(self, map_obj, layer, filename_generator):
        # cleanup temporary files
        for filename in filename_generator.filenames:
            vsi.unlink(filename)


class OutlinedBrowseLayerFactory(BrowseLayerMixIn, BaseMapServerLayerFactory):
    handled_layer_types = [OutlinedBrowseLayer]

    def create(self, map_obj, layer):
        filename_generator = FilenameGenerator('/vsimem/{uuid}.vrt')
        group_name = layer.name
        ranges = layer.ranges
        style = layer.style

        raster_style = (
            style if style and style in COLOR_SCALES else "blackwhite"
        )
        vector_style = (
            style if style and style in BASE_COLORS else "red"
        )

        generator = self.make_browse_layer_generator(
            map_obj, reversed(layer.browses), layer.map, filename_generator,
            group_name, ranges, raster_style
        )

        for browse, _ in generator:
            # create the outlines layer
            outlines_layer_obj = _create_polygon_layer(map_obj)
            shape_obj = ms.shapeObj.fromWKT(browse.footprint.wkt)
            outlines_layer_obj.addFeature(shape_obj)

            class_obj = _create_geometry_class(vector_style)
            outlines_layer_obj.insertClass(class_obj)

        return filename_generator

    def destroy(self, map_obj, layer, filename_generator):
        # cleanup temporary files
        for filename in filename_generator.filenames:
            vsi.unlink(filename)


class MaskLayerFactory(BaseMapServerLayerFactory):
    handled_layer_types = [MaskLayer]

    def create(self, map_obj, layer):
        layer_obj = _create_polygon_layer(map_obj)
        for mask in reversed(layer.masks):
            if mask.geometry:
                mask_geom = mask.geometry
            elif mask.filename:
                mask_geom = mask.load_geometry()
            else:
                continue

            shape_obj = ms.shapeObj.fromWKT(mask_geom.wkt)
            layer_obj.addFeature(shape_obj)

        layer_obj.insertClass(
            _create_geometry_class(layer.style or 'red', fill_opacity=1.0)
        )


class MaskedBrowseLayerFactory(BrowseLayerMixIn, BaseMapServerLayerFactory):
    handled_layer_types = [MaskedBrowseLayer]

    def create(self, map_obj, layer):
        filename_generator = FilenameGenerator('/vsimem/{uuid}.vrt')
        group_name = layer.name

        browses, masks = zip(*[
            [masked_browse.browse, masked_browse.mask]
            for masked_browse in reversed(layer.masked_browses)
        ]) if len(layer.masked_browses) > 0 else ([], [])

        generator = self.make_browse_layer_generator(
            map_obj, browses, layer.map, filename_generator,
            group_name, None, None
        )

        for browse_and_layer_objs, mask in zip(generator, masks):
            browse, browse_layer_objs = browse_and_layer_objs

            mask_name = 'mask__%d' % id(mask)

            # create mapserver layers for the mask
            mask_layer_obj = _create_polygon_layer(map_obj)
            mask_layer_obj.status = ms.MS_OFF
            mask_layer_obj.insertClass(
                _create_geometry_class("black", "white", fill_opacity=1.0)
            )

            if mask.geometry:
                mask_geom = mask.geometry
            elif mask.filename:
                mask_geom = mask.load_geometry()
            else:
                mask_geom = None

            # the current logic:
            # when dealing with validity masks:
            #  - when geometry is available, use it as a mask
            #  - when not, the browse shall be hidden
            # when dealing with 'normal' masks:
            #  - use footprint/image bounds and cutout geometry

            # TODO: currently it is assumed that all geometries
            # are in EPSG:4326 which is not necessarily the case.
            # Reprojection is required.

            outline = browse.footprint
            if mask_geom:
                if mask.validity:
                    outline = mask_geom
                else:
                    outline = outline - mask_geom
            elif mask.validity:
                outline = None

            if outline:
                shape_obj = ms.shapeObj.fromWKT(outline.wkt)
                mask_layer_obj.addFeature(shape_obj)

            mask_layer_obj.name = mask_name

            # set up the mapserver layers required for the browses
            for browse_layer_obj in browse_layer_objs:
                browse_layer_obj.mask = mask_name

        return filename_generator

    def destroy(self, map_obj, layer, filename_generator):
        # cleanup temporary files
        for filename in filename_generator.filenames:
            vsi.unlink(filename)


class OutlinesLayerFactory(BaseMapServerLayerFactory):
    handled_layer_types = [OutlinesLayer]

    def create(self, map_obj, layer):
        layer_obj = _create_polygon_layer(map_obj)
        footprint_masks = list(
            izip_longest(layer.footprints, layer.masks or [])
        )
        for footprint, mask in reversed(footprint_masks):
            if mask:
                if mask.geometry:
                    mask_geom = mask.geometry
                elif mask.filename:
                    mask_geom = mask.load_geometry()

                if mask.validity:
                    footprint = footprint.intersection(mask_geom)
                else:
                    footprint = footprint.difference(mask_geom)

            shape_obj = ms.shapeObj.fromWKT(footprint.wkt)
            layer_obj.addFeature(shape_obj)

        class_obj = _create_geometry_class(
            layer.style or 'red', fill_opacity=layer.fill
        )
        layer_obj.insertClass(class_obj)


# ------------------------------------------------------------------------------
# utils
# ------------------------------------------------------------------------------


def _create_raster_layer_objs(map_obj, extent, sr, data, filename_generator,
                              resample=None):
    layer_obj = ms.layerObj(map_obj)
    layer_obj.type = ms.MS_LAYER_RASTER
    layer_obj.status = ms.MS_ON

    layer_obj.data = data

    layer_obj.offsite = ms.colorObj(0, 0, 0)

    if extent:
        layer_obj.setMetaData("wms_extent", "%f %f %f %f" % extent)
        layer_obj.setExtent(*extent)

        if sr.srid is not None:
            short_epsg = "EPSG:%d" % sr.srid
            layer_obj.setMetaData("ows_srs", short_epsg)
            layer_obj.setMetaData("wms_srs", short_epsg)

    layer_obj.setProjection(sr.proj)

    if resample:
        layer_obj.setProcessingKey('RESAMPLE', resample)

    if extent and sr.srid and extent_crosses_dateline(extent, sr.srid):
        wrapped_extent = wrap_extent_around_dateline(extent, sr.srid)

        wrapped_layer_obj = ms.layerObj(map_obj)
        wrapped_layer_obj.type = ms.MS_LAYER_RASTER
        wrapped_layer_obj.status = ms.MS_ON

        wrapped_data = filename_generator.generate()
        vrt.with_extent(data, wrapped_extent, wrapped_data)
        wrapped_layer_obj.data = wrapped_data

        wrapped_layer_obj.offsite = ms.colorObj(0, 0, 0)

        wrapped_layer_obj.setMetaData("ows_srs", short_epsg)
        wrapped_layer_obj.setMetaData("wms_srs", short_epsg)
        wrapped_layer_obj.setProjection(sr.proj)

        wrapped_layer_obj.setExtent(*wrapped_extent)
        wrapped_layer_obj.setMetaData(
            "wms_extent", "%f %f %f %f" % wrapped_extent
        )
        return [layer_obj, wrapped_layer_obj]
    else:
        return [layer_obj]


def _create_polygon_layer(map_obj):
    layer_obj = ms.layerObj(map_obj)
    layer_obj.type = ms.MS_LAYER_POLYGON
    layer_obj.status = ms.MS_ON

    layer_obj.offsite = ms.colorObj(0, 0, 0)

    srid = 4326
    layer_obj.setProjection(crss.asProj4Str(srid))
    layer_obj.setMetaData("ows_srs", crss.asShortCode(srid))
    layer_obj.setMetaData("wms_srs", crss.asShortCode(srid))

    layer_obj.dump = True

    return layer_obj


def _create_geometry_class(color_name, background_color_name=None,
                           fill_opacity=None):
    cls_obj = ms.classObj()
    outline_style_obj = ms.styleObj()

    try:
        color = ms.colorObj(*BASE_COLORS[color_name])
    except KeyError:
        raise  # TODO

    outline_style_obj.outlinecolor = color
    cls_obj.insertStyle(outline_style_obj)

    if fill_opacity is not None:
        fill_style_obj = ms.styleObj()
        fill_style_obj.color = ms.colorObj(
            color.red, color.green, color.blue, int(255 * fill_opacity)
        )
        cls_obj.insertStyle(fill_style_obj)

    if background_color_name:
        outline_style_obj.backgroundcolor = ms.colorObj(
            *BASE_COLORS[background_color_name]
        )

    cls_obj.group = color_name
    return cls_obj


def _build_vrt(size, field_locations):
    path = join("/vsimem", uuid4().hex)
    size_x, size_y = size[:2]

    vrt_builder = vrt.VRTBuilder(size_x, size_y, vrt_filename=path)

    current = 1
    for field, location in field_locations:
        start = location.start_field
        end = location.end_field
        num = end - start + 1
        dst_band_indices = range(current, current + num)
        src_band_indices = range(1, num + 1)

        current += num

        for src_index, dst_index in zip(src_band_indices, dst_band_indices):
            vrt_builder.add_band(field.data_type)
            vrt_builder.add_simple_source(
                dst_index, location.path, src_index
            )

    del vrt_builder

    return path


def _create_raster_style(name, layer, minvalue=0, maxvalue=255,
                         nil_values=None):
    colors = COLOR_SCALES[name]

    nil_values = [float(nil_value) for nil_value in nil_values or []]

    if nil_values:
        offsite = ms.colorObj(*OFFSITE_COLORS.get(name, (0, 0, 0)))
        layer.offsite = offsite

        for nil_value in nil_values:
            cls = ms.classObj()
            cls.setExpression("([pixel] = %s)" % nil_value)
            cls.group = name

            style = ms.styleObj()
            style.color = offsite
            style.opacity = 0
            style.rangeitem = ""
            cls.insertStyle(style)
            layer.insertClass(cls)

    low_nil_values = [
        nil_value for nil_value in nil_values
        if nil_value <= minvalue
    ]
    high_nil_values = [
        nil_value for nil_value in nil_values
        if nil_value >= maxvalue
    ]

    # Create style for values below range, but make sure to not collide
    # with nil-values
    if low_nil_values:
        low_nil = max(low_nil_values)
        cls = ms.classObj()
        cls.setExpression(
            "([pixel] <= %s AND [pixel] > %s)" % (minvalue, low_nil)
        )
        cls.group = name
        style = ms.styleObj()
        style.color = ms.colorObj(*colors[0][1])
        cls.insertStyle(style)
        layer.insertClass(cls)
    else:
        cls = ms.classObj()
        cls.setExpression("([pixel] <= %s)" % (minvalue))
        cls.group = name
        style = ms.styleObj()
        style.color = ms.colorObj(*colors[0][1])
        cls.insertStyle(style)
        layer.insertClass(cls)

    interval = (maxvalue - minvalue)
    for prev_item, next_item in pairwise_iterative(colors):
        prev_perc, prev_color = prev_item
        next_perc, next_color = next_item

        cls = ms.classObj()
        cls.setExpression("([pixel] >= %s AND [pixel] < %s)" % (
            (minvalue + prev_perc * interval),
            (minvalue + next_perc * interval)
        ))
        cls.group = name

        style = ms.styleObj()
        style.mincolor = ms.colorObj(*prev_color)
        style.maxcolor = ms.colorObj(*next_color)
        style.minvalue = minvalue + prev_perc * interval
        style.maxvalue = minvalue + next_perc * interval
        style.rangeitem = ""
        cls.insertStyle(style)
        layer.insertClass(cls)

    # Create style for values above range, but make sure to not collide with
    # nil-values
    if high_nil_values:
        high_nil = min(high_nil_values)
        cls = ms.classObj()
        cls.setExpression(
            "([pixel] > %s AND [pixel] < %s)" % (maxvalue, high_nil)
        )
        cls.group = name
        style = ms.styleObj()
        style.color = ms.colorObj(*colors[0][1])
        cls.insertStyle(style)
        layer.insertClass(cls)
    else:
        cls = ms.classObj()
        cls.setExpression("([pixel] > %s)" % (maxvalue))
        cls.group = name
        style = ms.styleObj()
        style.color = ms.colorObj(*colors[-1][1])
        cls.insertStyle(style)
        layer.insertClass(cls)


def _get_range(field, range_=None):
    """ Gets the numeric range of a field
    """
    if range_:
        return tuple(range_)
    elif len(field.allowed_values) == 1:
        return field.allowed_values[0]
    elif field.data_type_range:
        return field.data_type_range
    return gdal.GDT_NUMERIC_LIMITS.get(field.data_type) or (0, 255)

# ------------------------------------------------------------------------------
# Layer factories
# ------------------------------------------------------------------------------


LAYER_FACTORIES = None


def _setup_factories():
    global LAYER_FACTORIES

    specifiers = getattr(
        settings, 'EOXS_MAPSERVER_LAYER_FACTORIES',
        DEFAULT_EOXS_MAPSERVER_LAYER_FACTORIES
    )
    LAYER_FACTORIES = [
        import_string(specifier)
        for specifier in specifiers
    ]


def get_layer_factories():
    if LAYER_FACTORIES is None:
        _setup_factories()
    return LAYER_FACTORIES



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.contrib.mapserver import Layer

from eoxserver.services.mapserver.wms.layerfactories.base import (
    AbstractLayerFactory, OffsiteColorMixIn
)


class CoverageBandsLayerFactory(OffsiteColorMixIn, AbstractLayerFactory):
    suffixes = ("_bands",)
    requires_connection = True

    def generate(self, eo_object, group_layer, suffix, options):
        name = eo_object.identifier + "_bands"
        layer = Layer(name)
        layer.setMetaData("ows_title", name)
        layer.setMetaData("wms_label", name)
        layer.addProcessing("CLOSE_CONNECTION=CLOSE")

        coverage = eo_object.cast()
        range_type = coverage.range_type

        req_bands = options["bands"]
        band_indices = []
        bands = []

        for req_band in req_bands:
            if isinstance(req_band, int):
                band_indices.append(req_band + 1)

                bands.append(range_type[req_band])
            else:
                for i, band in enumerate(range_type):
                    if band.name == req_band:
                        band_indices.append(i + 1)
                        bands.append(band)
                        break
                else:
                    raise Exception(
                        "Coverage '%s' does not have a band with name '%s'."
                        % (coverage.identifier, req_band)
                    )

        if len(req_bands) in (3, 4):
            indices_str = ",".join(map(str, band_indices))
            offsite_indices = list(map(lambda v: v-1, band_indices[:3]))
        elif len(req_bands) == 1:
            indices_str = ",".join(map(str, band_indices * 3))
            v = band_indices[0] - 1
            offsite_indices = [v, v, v]
        else:
            raise Exception("Invalid number of bands requested.")

        offsite = self.offsite_color_from_range_type(
            range_type, offsite_indices
        )
        options = self.get_render_options(coverage)
        self.set_render_options(layer, offsite, options)

        layer.setProcessingKey("BANDS", indices_str)

        if options.bands_scale_min and options.bands_scale_max:
            bands_scale_min = str(options.bands_scale_min).split(',')
            bands_scale_max = str(options.bands_scale_max).split(',')
            idx1, idx2, idx3 = offsite_indices
            layer.setProcessingKey("SCALE_1", "%s,%s" % (
                bands_scale_min[idx1], bands_scale_max[idx1]
            ))
            layer.setProcessingKey("SCALE_2", "%s,%s" % (
                bands_scale_min[idx2], bands_scale_max[idx2]
            ))
            layer.setProcessingKey("SCALE_3", "%s,%s" % (
                bands_scale_min[idx3], bands_scale_max[idx3]
            ))

        yield (layer, coverage.data_items.all())

    def generate_group(self, name):
        return Layer(name)



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2017 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.conf import settings
from django.utils.module_loading import import_string

from eoxserver.core.decoders import config


# default for EOXS_OPENSEARCH_FORMATS
DEFAULT_EOXS_OPENSEARCH_FORMATS = [
    'eoxserver.services.opensearch.formats.atom.AtomResultFormat',
    'eoxserver.services.opensearch.formats.rss.RSSResultFormat',
    'eoxserver.services.opensearch.formats.html.HTMLResultFormat',
    'eoxserver.services.opensearch.formats.kml.KMLResultFormat',
    'eoxserver.services.opensearch.formats.geojson.GeoJSONResultFormat',

]

# default for EOXS_OPENSEARCH_EXTENSIONS
DEFAULT_EOXS_OPENSEARCH_EXTENSIONS = [
    'eoxserver.services.opensearch.extensions.eo.EarthObservationExtension',
    'eoxserver.services.opensearch.extensions.geo.GeoExtension',
    'eoxserver.services.opensearch.extensions.time.TimeExtension',
    'eoxserver.services.opensearch.extensions.cql.CQLExtension',
]

# default for EOXS_OPENSEARCH_SUMMARY_TEMPLATE
DEFAULT_EOXS_OPENSEARCH_SUMMARY_TEMPLATE = "opensearch/summary.html"

# default for the EOXS_OPENSEARCH_RECORD_MODEL
DEFAULT_EOXS_OPENSEARCH_RECORD_MODEL = "eoxserver.resources.coverages.models.EOObject"


def get_opensearch_record_model():
    class_name = getattr(
        settings, 'EOXS_OPENSEARCH_RECORD_MODEL', DEFAULT_EOXS_OPENSEARCH_RECORD_MODEL
    )
    return import_string(class_name)


class OpenSearchConfigReader(config.Reader):
    section = "services.opensearch"
    default_count = config.Option(type=int, default=100)
    max_count = config.Option(type=int, default=200)



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


from eoxserver.core import implements
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import xml
from eoxserver.contrib import mapserver as ms
from eoxserver.resources.coverages import models
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.services.mapserver.wms.util import MapServerWMSBaseComponent
from eoxserver.services.ows.wms.interfaces import (
    WMSFeatureInfoRendererInterface
)
from eoxserver.services.ows.wcs.v20.encoders import WCS20EOXMLEncoder
from eoxserver.services.result import (
    result_set_from_raw_data, get_content_type, ResultBuffer
)
from eoxserver.services.urls import get_http_service_url

class MapServerWMSFeatureInfoRenderer(MapServerWMSBaseComponent):
    """ A WMS feature info renderer using MapServer.
    """
    implements(WMSFeatureInfoRendererInterface)

    
    def render(self, layer_groups, request_values, request, **options):
        config = CapabilitiesConfigReader(get_eoxserver_config())
        http_service_url = get_http_service_url(request)
        map_ = ms.Map()
        map_.setMetaData({
            "enable_request": "*",
            "onlineresource": http_service_url,
        }, namespace="ows")

        map_.setMetaData("wms_getfeatureinfo_formatlist", "text/html")
        map_.setProjection("EPSG:4326")

        session = self.setup_map(layer_groups, map_, options)

        # check if the required format is EO O&M
        frmt = pop_param(request_values, "info_format")
        use_eoom = False
        if frmt in ("application/xml", "text/xml"):
            request_values.append(("info_format", "application/vnd.ogc.gml"))
            use_eoom = True
        else:
            request_values.append(("info_format", frmt))
        
        with session:
            request = ms.create_request(request_values)
            raw_result = map_.dispatch(request)
            result = result_set_from_raw_data(raw_result)

            if not use_eoom:
                # just return the response
                return result, get_content_type(result)
            else:
                # do a postprocessing step and get all identifiers in order
                # to encode them with EO O&M
                decoder = GMLFeatureDecoder(result[0].data_file.read())
                identifiers = decoder.identifiers
                coverages = models.Coverage.objects.filter(
                    identifier__in=identifiers
                )

                # sort the result with the returned order of coverages
                lookup_table = dict((c.identifier, c) for c in coverages)
                coverages = [
                    lookup_table[identifier] for identifier in identifiers
                ]

                # encode the coverages with the EO O&M 
                encoder = WCS20EOXMLEncoder()

                return [
                    ResultBuffer(
                        encoder.serialize(
                            encoder.encode_coverage_descriptions(coverages)
                        ), encoder.content_type
                    )
                ], encoder.content_type



def pop_param(request_values, name, default=None):
    """ Helper to pop one param from a key-value list
    """
    for param_name, value in request_values:
        if param_name.lower() == name:
            request_values.remove((param_name, value))
            return value
    return default


class GMLFeatureDecoder(xml.Decoder):
    identifiers = xml.Parameter("//identifier/text()", num="*")



#-------------------------------------------------------------------------------
#
#  Input objects used by the execute requests and responses
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from .base import ParamMetadata

class InputReference(ParamMetadata):
    # pylint: disable=too-few-public-methods, too-many-arguments
    """ Input data reference class.

    Constructor parameters:
        href         input reference URL
        identifier   input item identifier
        title        user defined title
        abstract     user defined abstract
        headers      additional HTTP request headers
        body         optional HTTP/POST request payload
        method       reference method ('GET' or 'POST')
        mime_type    reference ComplexData mime-type
        encoding     reference ComplexData encoding
        schema       reference ComplexData schema
        body_href    optional HTTP/POST request payload reference URL
    """

    def __init__(self, href, identifier, title=None, abstract=None,
                 headers=None, body=None, method=None, mime_type=None,
                 encoding=None, schema=None, body_href=None):
        ParamMetadata.__init__(
            self, identifier, title, abstract, None, None,
            mime_type, encoding, schema
        )
        self.href = href
        self.headers = headers or ()
        self.body = body
        self.body_href = body_href
        self.method = method


class InputData(ParamMetadata):
    # pylint: disable=too-few-public-methods, too-many-arguments
    """ Generic container for the raw data inputs.  An instances of this class
    holds the inputs as decoded from various WPS requests before their
    validation and conversion to their configured data-type.

    Constructor parameters:
        data         unparsed (raw) data payload (byte string)
        identifier   input item identifier
        title        user defined title
        abstract     user defined abstract
        uom          input LiteralData UOM
        crs          input BoundingBoxData CRS
        mime_type    input ComplexData mime-type
        encoding     input ComplexData encoding
        schema       input ComplexData schema
        asurl        indicates whether the decoded input comes from
                     a URL encoded request (KVP) or not.
    """
    def __init__(self, data, identifier, title=None, abstract=None,
                 uom=None, crs=None, mime_type=None,
                 encoding=None, schema=None, asurl=False):
        ParamMetadata.__init__(
            self, identifier, title, abstract, uom, crs,
            mime_type, encoding, schema
        )
        self.data = data
        self.asurl = asurl



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Stephan Krause <stephan.krause@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2011 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import ctypes as C
from ctypes.util import find_library
import logging
try:
    from itertools import chain, izip
except ImportError :
    from itertools import chain
    izip = zip

import math

from eoxserver.contrib import gdal, osr
from eoxserver.core.util.rect import Rect
from eoxserver.core.util.xmltools import parse, etree
from eoxserver.contrib import vsi
from django.utils.six import string_types, b

try:
    # Python 2
    xrange
except NameError:
    # Python 3, xrange is now named range
    xrange = range
#-------------------------------------------------------------------------------
# approximation transformer's threshold in pixel units
# 0.125 is the default value used by CLI gdalwarp tool


APPROX_ERR_TOL = 0.125

#-------------------------------------------------------------------------------
# GDAL transfomer methods

METHOD_GCP = 1
METHOD_TPS = 2
METHOD_TPS_LSQ = 3

METHOD2STR = {
    METHOD_GCP: "METHOD_GCP",
    METHOD_TPS: "METHOD_TPS",
    METHOD_TPS_LSQ: "METHOD_TPS_LSQ"
}

#-------------------------------------------------------------------------------

logger = logging.getLogger(__name__)


class WARP_OPTIONS(C.Structure):
    _fields_ = [
        ("papszWarpOptions", C.POINTER(C.c_char_p)),
        ("dfWarpMemoryLimit", C.c_double),
        ("eResampleAlg", C.c_int),
        ("eWorkingDataType", C.c_int),
        ("hSrcDS", C.c_void_p),
        ("hDstDS", C.c_void_p),
        ("nBandCount", C.c_int),
        ("panSrcBands", C.POINTER(C.c_int)),
        ("panDstBands", C.POINTER(C.c_int)),
        ("nSrcAlphaBand", C.c_int),
        ("nDstAlphaBand", C.c_int),
        ("padfSrcNoDataReal", C.POINTER(C.c_double)),
        ("padfSrcNoDataImag", C.POINTER(C.c_double)),
        ("padfDstNoDataReal", C.POINTER(C.c_double)),
        ("padfDstNoDataImag", C.POINTER(C.c_double)),
        ("pfnProgress", C.c_void_p),
        ("pProgressArg", C.c_void_p),
        ("pfnTransformer", C.c_void_p),
        ("pTransformerArg", C.c_void_p),
        ("papfnSrcPerBandValidityMaskFunc", C.c_void_p),
        ("papSrcPerBandValidityMaskFuncArg", C.c_void_p),
        ("pfnSrcValidityMaskFunc", C.c_void_p),
        ("pSrcValidityMaskFuncArg", C.c_void_p),
        ("pfnSrcDensityMaskFunc", C.c_void_p),
        ("pSrcDensityMaskFuncArg", C.c_void_p),
        ("pfnDstDensityMaskFunc", C.c_void_p),
        ("pDstDensityMaskFuncArg", C.c_void_p),
        ("pfnDstValidityMaskFunc", C.c_void_p),
        ("pDstValidityMaskFuncArg", C.c_void_p),
        ("pfnPreWarpChunkProcessor", C.c_void_p),
        ("pPreWarpProcessorArg", C.c_void_p),
        ("pfnPostWarpChunkProcessor", C.c_void_p),
        ("pPostWarpProcessorArg", C.c_void_p),
        ("hCutline", C.c_void_p),
        ("dfCutlineBlendDist", C.c_double),
    ]

_libgdal = C.LibraryLoader(C.CDLL).LoadLibrary(find_library("gdal"))


class GCP(C.Structure):
    _fields_ = [
        ("pszId", C.c_char_p),
        ("pszInfo", C.c_char_p),
        ("dfGCPPixel", C.c_double),
        ("dfGCPLine", C.c_double),
        ("dfGCPX", C.c_double),
        ("dfGCPY", C.c_double),
        ("dfGCPZ", C.c_double),
    ]

GDALGetGCPs = _libgdal.GDALGetGCPs
GDALGetGCPs.restype = C.c_void_p # actually array of GCPs, but more info not required
GDALGetGCPs.argtypes = [C.c_void_p]


# baseline GDAL transformer creation functions

GDALCreateGCPTransformer = _libgdal.GDALCreateGCPTransformer
GDALCreateGCPTransformer.restype = C.c_void_p
# TODO: argtypes

GDALCreateTPSTransformer = _libgdal.GDALCreateTPSTransformer
GDALCreateTPSTransformer.restype = C.c_void_p
# TODO: argtypes

GDALGenImgProjTransform = _libgdal.GDALGenImgProjTransform

GDALCreateGenImgProjTransformer2 =_libgdal.GDALCreateGenImgProjTransformer2
GDALCreateGenImgProjTransformer2.restype = C.c_void_p
GDALCreateGenImgProjTransformer2.argtypes = [C.c_void_p, C.c_void_p, C.POINTER(C.c_char_p)]

GDALUseTransformer = _libgdal.GDALUseTransformer
GDALUseTransformer.restype = C.c_int
GDALUseTransformer.argtypes = [C.c_void_p, C.c_int, C.c_int, C.POINTER(C.c_double), C.POINTER(C.c_double), C.POINTER(C.c_double), C.POINTER(C.c_int)]

GDALDestroyTransformer = _libgdal.GDALDestroyTransformer
GDALDestroyTransformer.argtypes = [C.c_void_p]

# extended GDAL transformer creation functions
try:
    GDALCreateTPS2TransformerExt = _libgdal.GDALCreateTPS2TransformerExt
    GDALCreateTPS2TransformerExt.restype = C.c_void_p
    # TODO: argtypes
except AttributeError:
    GDALCreateTPS2TransformerExt = None

try:
    GDALCreateTPS2TransformerLSQGrid = _libgdal.GDALCreateTPS2TransformerLSQGrid
    GDALCreateTPS2TransformerLSQGrid.restype = C.c_void_p
    # TODO: argtypes
except AttributeError:
    GDALCreateTPS2TransformerLSQGrid = None


OCTNewCoordinateTransformation = _libgdal.OCTNewCoordinateTransformation
OCTNewCoordinateTransformation.restype = C.c_void_p
OCTNewCoordinateTransformation.argtypes = [C.c_void_p, C.c_void_p]
#OCTNewCoordinateTransformation.errcheck = None # TODO!

OCTDestroyCoordinateTransformation = _libgdal.OCTDestroyCoordinateTransformation
OCTDestroyCoordinateTransformation.argtypes = [C.c_void_p]

OCTTransform = _libgdal.OCTTransform
OCTTransform.argtypes = [C.c_void_p, C.c_int, C.POINTER(C.c_double), C.POINTER(C.c_double), C.POINTER(C.c_double)]

GDALCreateWarpOptions = _libgdal.GDALCreateWarpOptions
GDALCreateWarpOptions.restype = C.POINTER(WARP_OPTIONS)

GDALSuggestedWarpOutput = _libgdal.GDALSuggestedWarpOutput
GDALSuggestedWarpOutput.restype = C.c_int
GDALSuggestedWarpOutput.argtypes = [C.c_void_p, C.c_void_p, C.c_void_p, C.c_double * 6, C.POINTER(C.c_int), C.POINTER(C.c_int)]

GDALSetGenImgProjTransformerDstGeoTransform = _libgdal.GDALSetGenImgProjTransformerDstGeoTransform
GDALSetGenImgProjTransformerDstGeoTransform.argtypes = [C.c_void_p, C.c_double * 6]

GDALCreateWarpedVRT = _libgdal.GDALCreateWarpedVRT
GDALCreateWarpedVRT.restype = C.c_void_p
GDALCreateWarpedVRT.argtypes = [C.c_void_p, C.c_int, C.c_int, C.c_double * 6, C.POINTER(WARP_OPTIONS)]

GDALAutoCreateWarpedVRT = _libgdal.GDALAutoCreateWarpedVRT
GDALAutoCreateWarpedVRT.restype = C.c_void_p
GDALAutoCreateWarpedVRT.argtypes = [C.c_void_p, C.c_char_p, C.c_char_p, C.c_int, C.c_double, C.POINTER(WARP_OPTIONS)]


GDALCreateApproxTransformer = _libgdal.GDALCreateApproxTransformer
GDALCreateApproxTransformer.restype = C.c_void_p
GDALCreateApproxTransformer.argtypes = [C.c_void_p, C.c_void_p, C.c_double]

GDALApproxTransformerOwnsSubtransformer = _libgdal.GDALApproxTransformerOwnsSubtransformer
GDALApproxTransformerOwnsSubtransformer.argtypes = [C.c_void_p, C.c_bool]

GDALSetDescription = _libgdal.GDALSetDescription
GDALSetDescription.argtypes = [C.c_void_p, C.c_char_p]


GDALSetProjection = _libgdal.GDALSetProjection
GDALSetProjection.argtypes = [C.c_void_p, C.c_char_p]

GDALDestroyWarpOptions = _libgdal.GDALDestroyWarpOptions
GDALDestroyWarpOptions.argtypes = [C.POINTER(WARP_OPTIONS)]

GDALClose = _libgdal.GDALClose
GDALClose.argtypes = [C.c_void_p]


class Transformer(object):
    def __init__(self, handle):
        self._handle = handle

    @property
    def _as_parameter_(self):
        return self._handle

    def __del__(self):
        GDALDestroyTransformer(self._handle)

    #def __call__(self, points, ):


class CoordinateTransformation(object):

    def __init__(self, src_srs, dst_srs):
        self._handle = OCTNewCoordinateTransformation(
            C.cast(int(src_srs.this), C.c_void_p),
            C.cast(int(dst_srs.this), C.c_void_p)
        )

    @property
    def _as_parameter_(self):
        return self._handle

    def __del__(self):
        OCTDestroyCoordinateTransformation(self)


def _create_referenceable_grid_transformer(ds, method, order):
    # TODO: check method and order
    num_gcps = ds.GetGCPCount()
    gcps = GDALGetGCPs(C.cast(int(ds.this), C.c_void_p))
    handle = None
    if method == METHOD_GCP:
        handle = GDALCreateGCPTransformer(num_gcps, C.c_void_p(gcps), order, 0)
    elif method == METHOD_TPS:
        if GDALCreateTPS2TransformerExt:
            handle = GDALCreateTPS2TransformerExt(num_gcps, C.c_void_p(gcps), 0, order)
        else:
            handle = GDALCreateTPSTransformer(num_gcps, gcps, 0)

    elif method == METHOD_TPS_LSQ and GDALCreateTPS2TransformerLSQGrid:
        handle = GDALCreateTPS2TransformerLSQGrid(num_gcps, gcps, 0, order, 0, 0)
    elif method == METHOD_TPS_LSQ:
        raise AttributeError("GDALCreateTPS2TransformerLSQGrid not available")

    else:
        raise

    return Transformer(handle)


CSLFetchNameValue = _libgdal.CSLFetchNameValue
CSLFetchNameValue.restype = C.c_char_p
CSLFetchNameValue.argtypes = [C.POINTER(C.c_char_p), C.c_char_p]

CSLSetNameValue = _libgdal.CSLSetNameValue
CSLSetNameValue.restype = C.POINTER(C.c_char_p)
CSLSetNameValue.argtypes = [C.POINTER(C.c_char_p), C.c_char_p, C.c_char_p]

CSLDestroy = _libgdal.CSLDestroy
CSLDestroy.argtypes = [C.POINTER(C.c_char_p)]

CSLCount = _libgdal.CSLCount
CSLCount.restype = C.c_int
CSLCount.argtypes = [C.POINTER(C.c_char_p)]

CPLParseNameValue = _libgdal.CPLParseNameValue
CPLParseNameValue.restype = C.c_char_p
CPLParseNameValue.argtypes = [C.c_char_p, C.POINTER(C.c_char_p)]

CPLMalloc = _libgdal.CPLMalloc
CPLMalloc.restype = C.c_void_p

CPLFree = _libgdal.free
CPLFree.argtypes = [C.c_void_p]


class CSL(object):
    """ Wrapper for GDAL CSL API.
    """

    def __init__(self, **kwargs):
        self._handle = None
        for key, value in kwargs.items():
            self[key] = value

    @property
    def _as_parameter_(self):
        if not self._handle:
            return C.cast(self._handle, C.POINTER(C.c_char_p))
        return self._handle

    def __getitem__(self, key):
        value = CSLFetchNameValue(self, key)
        if not value:
            raise KeyError(key)

    def __setitem__(self, key, value):
        self._handle = CSLSetNameValue(self, key, value)

    def __del__(self):
        CSLDestroy(self)

    def __len__(self):
        return CSLCount(self)

    def __iter__(self):
        if not self._handle:
            raise StopIteration

        i = 0
        while True:
            p = self._handle[i]
            if not p:
                break
            key, _, value = p.partition("=")
            yield key, value
            i += 1

    def __repr__(self):
        return "{%s}" % (
            ", ".join(
                "%r: %r" % (key, value) for key, value in self
            )
        )


def _create_generic_transformer(src_ds, src_wkt, dst_ds, dst_wkt, method, order):
    # TODO: check method and order

    try:
        src_ds = C.c_void_p(int(src_ds.this))
    except AttributeError:
        pass
    try:
        dst_ds = C.c_void_p(int(dst_ds.this))
    except AttributeError:
        pass

    options = CSL()

    if src_wkt:
        options["SRC_SRS"] = src_wkt
    if dst_wkt:
        options["DST_SRS"] = dst_wkt

    if method == METHOD_GCP:
        options["METHOD"] = "GCP_POLYNOMIAL"
        options["GCPS_OK"] = "TRUE"
        if order > 0:
            options["MAX_GCP_ORDER"] = str(order)

    elif method in (METHOD_TPS, METHOD_TPS_LSQ):
        if GDALCreateTPS2TransformerLSQGrid:
            options["METHOD"] = "GCP_TPS2"
            options["TPS2_AP_ORDER"] = str(order)
            if method == METHOD_TPS_LSQ:
                options["TPS2_LSQ_GRID"] = "1"
                options["TPS2_LSQ_GRID_NX"] = "0"
                options["TPS2_LSQ_GRID_NY"] = "0"
        else:
            options["METHOD"] = "GCP_TPS"

    else:
        raise RuntimeError("Unknown transformation method.")

    # TODO: proper error handling
    handle = GDALCreateGenImgProjTransformer2(src_ds, dst_ds, options)
    if not handle:
        raise Exception(gdal.GetLastErrorMsg())

    transformer = Transformer(handle)
    return transformer


def get_footprint_wkt(ds, method=METHOD_GCP, order=0):
    """ Returns the footprint of the GDAL Dataset using its GCPs for
    calculation.


    :param method: either of :const:`METHOD_GCP`, :const:`METHOD_TPS` or
                   :const:`METHOD_TPS_LSQ`.

    :param order: (method specific):

                  - GCP (order of global fitting polynomial)
                      - ``0`` for automatic order
                      - ``1``, ``2``, and ``3`` for 1st, 2nd and 3rd polynomial
                      - order

                  - TPS and TPS_LSQ (order of augmenting polynomial)
                      - ``-1``  for no-polynomial augmentation
                      - ``0``  for 0th order (constant offset)
                      - ``1``, ``2``, and ``3`` for 1st, 2nd and 3rd polynomial
                        order

    General guide:

        - method TPS, order 3 should work in most cases
        - method TPS_LSQ, order 3 should work in cases of an excessive number
          of tiepoints but it may become wobbly for small number of tiepoints

       The global polynomial (GCP) interpolation does not work well for images
       covering large geographic areas (e.g., ENVISAT ASAR and MERIS).

    .. note:: The default parameters are left for backward compatibility.
          They can be, however, often inappropriate!
    """
    transformer = _create_referenceable_grid_transformer(ds, method, order)

    x_size = ds.RasterXSize
    y_size = ds.RasterYSize

    x_e = max(x_size // 100 - 1, 0)
    y_e = max(y_size // 100 - 1, 0)

    num_points = 4 + 2 * x_e + 2 * y_e
    coord_array_type = (C.c_double * num_points)
    x = coord_array_type()
    y = coord_array_type()
    z = coord_array_type()

    success = (C.c_int * num_points)()

    for i in xrange(1, x_e + 1):
        x[i] = float(i * x_size / x_e)
        y[i] = 0.0

    x[x_e + 1] = x_size

    for i in xrange(1, y_e + 1):
        x[x_e + 1 + i] = float(x_size)
        y[x_e + 1 + i] = float(i * y_size / y_e)

    x[x_e + y_e + 2] = x_size
    y[x_e + y_e + 2] = y_size

    for i in xrange(1, x_e + 1):
        x[x_e + y_e + 2 + i] = float(x_size - i * x_size / x_e)
        y[x_e + y_e + 2 + i] = y_size

    y[x_e * 2 + y_e + 3] = y_size

    for i in xrange(1, y_e + 1):
        x[x_e * 2 + y_e + 3 + i] = 0.0
        y[x_e * 2 + y_e + 3 + i] = float(y_size - i * y_size / y_e)

    GDALUseTransformer(transformer, False, num_points, x, y, z, success)

    return "POLYGON((%s))" % (
        ",".join(
            "%f %f" % (coord_x, coord_y)
            for coord_x, coord_y in chain(izip(x, y), ((x[0], y[0]),))
        )
    )


def rect_from_subset(path_or_ds, srid, minx, miny, maxx, maxy,
                     method=METHOD_GCP, order=0):
    """ Returns the smallest area of an image for the given spatial subset.

    :param path_or_ds: a :class:`GDAL Dataset <eoxserver.contrib.gdal.Dataset>`
                       or a path to such
    :param srid: the SRID the ``minx``, ``miny``, ``maxx`` and ``maxy`` are
                 expressed in
    :param minx: the minimum X subset coordinate
    :param miny: the minimum Y subset coordinate
    :param maxx: the maximum X subset coordinate
    :param maxy: the maximum Y subset coordinate
    :param method: either of :const:`METHOD_GCP`, :const:`METHOD_TPS` or
                   :const:`METHOD_TPS_LSQ`.
    :param order: the order of the function; see :func:`get_footprint_wkt` for
                  reference
    :returns: a :class:`Rect <eoxserver.core.util.rect.Rect>` portraing the
              subset in image coordinates
    """

    ds = path_or_ds

    x_size = ds.RasterXSize
    y_size = ds.RasterYSize

    transformer = _create_referenceable_grid_transformer(ds, method, order)

    gcp_srs = osr.SpatialReference(ds.GetGCPProjection())

    subset_srs = osr.SpatialReference()
    subset_srs.ImportFromEPSG(srid)

    coord_array_type = (C.c_double * 4)
    x = coord_array_type()
    y = coord_array_type()
    z = coord_array_type()

    success = (C.c_int * 4)()

    x[1] = float(x_size)
    y[1] = 0.0

    x[2] = float(x_size)
    y[2] = float(y_size)

    x[3] = 0.0
    y[3] = float(y_size)

    GDALUseTransformer(transformer, False, 4, x, y, z, success)

    dist = min(
        (max(x) - min(x)) / (x_size / 100),
        (max(y) - min(y)) / (y_size / 100)
    )

    x[0] = x[3] = minx
    x[1] = x[2] = maxx
    y[0] = y[1] = miny
    y[2] = y[3] = maxy

    ct = CoordinateTransformation(subset_srs, gcp_srs)

    OCTTransform(ct, 4, x, y, z)

    num_x = int(math.ceil((max(x) - min(x)) / dist))
    num_y = int(math.ceil((max(y) - min(y)) / dist))

    x_step = (maxx - minx) / num_x
    y_step = (maxy - miny) / num_y

    num_points = 4 + 2 * num_x + 2 * num_y

    coord_array_type = (C.c_double * num_points)
    x = coord_array_type()
    y = coord_array_type()
    z = coord_array_type()
    success = (C.c_int * num_points)()

    x[0] = minx
    y[0] = miny

    for i in xrange(1, num_x + 1):
        x[i] = minx + i * x_step
        y[i] = miny

    x[num_x + 1] = maxx
    y[num_x + 1] = miny

    for i in xrange(1, num_y + 1):
        x[num_x + 1 + i] = maxx
        y[num_x + 1 + i] = miny + i * y_step

    x[num_x + num_y + 2] = maxx
    y[num_x + num_y + 2] = maxy

    for i in xrange(1, num_x + 1):
        x[num_x + num_y + 2 + i] = maxx - i * x_step
        y[num_x + num_y + 2 + i] = maxy

    x[num_x * 2 + num_y + 3] = minx
    y[num_x * 2 + num_y + 3] = maxy

    for i in xrange(1, num_y + 1):
        x[num_x * 2 + num_y + 3 + i] = minx
        y[num_x * 2 + num_y + 3 + i] = maxy - i * y_step

    OCTTransform(ct, num_points, x, y, z)
    GDALUseTransformer(transformer, True, num_points, x, y, z, success)

    minx = int(math.floor(min(x)))
    miny = int(math.floor(min(y)))
    size_x = int(math.ceil(max(x) - minx) + 1)
    size_y = int(math.ceil(max(y) - miny) + 1)

    return Rect(minx, miny, size_x, size_y)


def create_rectified_vrt(path_or_ds, vrt_path, srid_or_wkt=None,
                         resample=0, memory_limit=0.0,
                         max_error=APPROX_ERR_TOL, method=METHOD_GCP, order=0,
                         size=None, resolution=None):

    """ Creates a VRT dataset that symbolizes a rectified version of the
    passed "referenceable" GDAL dataset.

    :param path_or_ds: a :class:`GDAL Dataset <eoxserver.contrib.gdal.Dataset>`
                       or a path to such
    :param vrt_path: the path to store the VRT dataset under

    :param resample: the resample method to be used; defaults to 0 which means
                     a nearest neighbour resampling
    :param memory_limit: the memory limit; by default no limit is used
    :param max_error: the maximum allowed error
    :param method: either of :const:`METHOD_GCP`, :const:`METHOD_TPS` or
                   :const:`METHOD_TPS_LSQ`.
    :param order: the order of the function; see :func:`get_footprint_wkt` for
                  reference
    """

    if size and resolution:
        raise ValueError('size and resolution ar mutually exclusive')

    ds = _open_ds(path_or_ds)
    ptr = C.c_void_p(int(ds.this))

    if isinstance(srid_or_wkt, int):
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(srid_or_wkt)
        wkt = srs.ExportToWkt()
        srs = None
    elif isinstance(srid_or_wkt, str):
        wkt = srid_or_wkt
    else:
        wkt = ds.GetGCPProjection()

    # transformer = _create_generic_transformer(
    #     ds, None, None, wkt, method, order
    # )

    # x_size = C.c_int()
    # y_size = C.c_int()
    # geotransform = (C.c_double * 6)()

    # GDALSuggestedWarpOutput(
    #     ptr,
    #     GDALGenImgProjTransform, transformer, geotransform,
    #     C.byref(x_size), C.byref(y_size)
    # )

    # GDALSetGenImgProjTransformerDstGeoTransform(transformer, geotransform)

    # options = GDALCreateWarpOptions()
    # options.dfWarpMemoryLimit = memory_limit
    # options.eResampleAlg = resample
    # options.pfnTransformer = GDALGenImgProjTransform
    # options.pTransformerArg = transformer
    # options.hDstDS = C.c_void_p(int(ds.this))

    # nb = options.nBandCount = ds.RasterCount

    # src_bands = C.cast(CPLMalloc(C.sizeof(C.c_int) * nb), C.POINTER(C.c_int))
    # dst_bands = C.cast(CPLMalloc(C.sizeof(C.c_int) * nb), C.POINTER(C.c_int))

    # # ctypes.cast(x, ctypes.POINTER(ctypes.c_ulong))

    # options.panSrcBands = src_bands
    # options.panDstBands = dst_bands

    # # TODO: nodata value setup
    # for i in xrange(nb):
    #     options.panSrcBands[i] = i + 1
    #     options.panDstBands[i] = i + 1

    # if max_error > 0:
    #     GDALApproxTransform = _libgdal.GDALApproxTransform

    #     options.pTransformerArg = GDALCreateApproxTransformer(
    #         options.pfnTransformer, options.pTransformerArg, max_error
    #     )
    #     options.pfnTransformer = GDALApproxTransform
        # TODO: correct for python
        #GDALApproxTransformerOwnsSubtransformer(options.pTransformerArg, False)

    # if size:
    #     extent = _to_extent(x_size.value, y_size.value, geotransform)
    #     size_x, size_y = size
    #     x_size.value = size_x
    #     y_size.value = size_y
    #     geotransform = _to_gt(size[0], size[1], extent)

    # elif resolution:
    #     extent = _to_extent(x_size.value, y_size.value, geotransform)

    #     geotransform[1] = resolution[0]
    #     geotransform[5] = resolution[1]

    #     size_x, size_y = _to_size(geotransform, extent)
    #     x_size.value = size_x
    #     y_size.value = size_y

    # vrt_ds = GDALCreateWarpedVRT(ptr, x_size, y_size, geotransform, options)
    if isinstance(wkt, str):
        wkt = b(wkt)
    vrt_ds = GDALAutoCreateWarpedVRT(ptr, None, wkt, resample, max_error, None)
    # GDALSetProjection(vrt_ds, wkt)
    if isinstance(vrt_path, str):
       vrt_path = b(vrt_path)

    GDALSetDescription(vrt_ds, vrt_path)
    GDALClose(vrt_ds)
    # GDALDestroyWarpOptions(options)

    # if size of resolution is overridden parse the VRT and adjust settings
    if size or resolution:
        with vsi.open(vrt_path) as f:
            root = parse(f).getroot()

        size_x = int(root.attrib['rasterXSize'])
        size_y = int(root.attrib['rasterYSize'])
        gt_elem = root.find('GeoTransform')

        gt = [
            float(value.strip())
            for value in gt_elem.text.strip().split(',')
        ]

        if size:
            extent = _to_extent(size_x, size_y, gt)
            size_x, size_y = size
            gt = _to_gt(size[0], size[1], extent)

        elif resolution:
            extent = _to_extent(size_x, size_y, gt)

            gt[1] = resolution[0]
            gt[5] = resolution[1]

            size_x, size_y = _to_size(gt, extent)

        # Adjust XML
        root.attrib['rasterXSize'] = str(size_x)
        root.attrib['rasterYSize'] = str(size_y)

        gt_str = ",".join(str(v) for v in gt)
        gt_elem.text = gt_str
        root.find(
            'GDALWarpOptions/Transformer/ApproxTransformer/'
            'BaseTransformer/GenImgProjTransformer/DstGeoTransform'
        ).text = gt_str

        inv_gt = gdal.InvGeoTransform(gt)
        root.find(
            'GDALWarpOptions/Transformer/ApproxTransformer/'
            'BaseTransformer/GenImgProjTransformer/DstInvGeoTransform'
        ).text = ",".join(str(v) for v in inv_gt)

        # write XML back to file
        with vsi.open(vrt_path, "w") as f:
            f.write(etree.tostring(root, pretty_print=True))


def _to_extent(size_x, size_y, gt):
    x_a = gt[0]
    x_b = gt[0] + gt[1] * size_x
    y_a = gt[3]
    y_b = gt[3] + gt[5] * size_y

    return (min(x_a, x_b), min(y_a, y_b), max(x_a, x_b), max(y_a, y_b))


def _to_gt(size_x, size_y, extent):
    ex = extent[2] - extent[0]
    ey = extent[3] - extent[1]
    return [
        extent[0],
        ex / float(size_x),
        0.0,
        extent[3],
        0.0,
        ey / float(size_y) * -1
    ]


def _to_size(gt, extent):
    dx = abs(gt[1])
    dy = abs(gt[5])

    ex = extent[2] - extent[0]
    ey = extent[3] - extent[1]

    return int(ex / dx), int(ey / dy)




def suggested_warp_output(ds, src_wkt, dst_wkt, method=METHOD_GCP, order=0):
    """ Returns the suggested size and geotransform tuple for the given dataset.

    :param ds: a :class:`GDAL Dataset <eoxserver.contrib.gdal.Dataset>`
    :param src_wkt: override the source WKT notation of the transform
    :param dst_wkt: the destination projection in WKT notation
    :param method: either of :const:`METHOD_GCP`, :const:`METHOD_TPS` or
                   :const:`METHOD_TPS_LSQ`.
    :param order: the order of the function; see :func:`get_footprint_wkt` for
                  reference
    :returns: the X size of the suggested warped image
    :returns: the Y size of the suggested warped image
    :returns: the geotransform tuple of the suggested warped image
    """
    geotransform = (C.c_double * 6)()
    x_size = C.c_int()
    y_size = C.c_int()
    transformer = _create_generic_transformer(
        ds, src_wkt, dst_wkt, method, order
    )
    GDALSuggestedWarpOutput(
        ds, GDALGenImgProjTransform, transformer,
        geotransform, C.byref(x_size), C.byref(y_size)
    )

    return x_size.value, y_size.value, tuple(geotransform)


def reproject_image(src_ds, src_wkt, dst_ds, dst_wkt,
                    resample=0, memory_limit=0.0,
                    max_error=APPROX_ERR_TOL, method=METHOD_GCP, order=0):
    """ Transforms the source dataset to the destination dataset.

    :param src_ds: the source :class:`GDAL Dataset
                   <eoxserver.contrib.gdal.Dataset>`
    :param src_wkt: override the source projection in WKT notation
    :param dst_ds: the source :class:`GDAL Dataset
                   <eoxserver.contrib.gdal.Dataset>`
    :param dst_wkt: override the destination projection in WKT notation
    :param resample: the resample method to be used; defaults to 0 which means
                     a nearest neighbour resampling
    :param memory_limit: the memory limit; by default no limit is used
    :param max_error: the maximum allowed error
    :param method: either of :const:`METHOD_GCP`, :const:`METHOD_TPS` or
                   :const:`METHOD_TPS_LSQ`.
    :param order: the order of the function; see :func:`get_footprint_wkt` for
                  reference
    """

    transformer = _create_generic_transformer(
        src_ds.this, src_wkt, dst_ds.this, dst_wkt, method, order
    )
    size_x = dst_ds.RasterXSize
    size_y = dst_ds.RasterYSize

    options = GDALCreateWarpOptions()
    options.eResampleAlg = resample
    options.dfWarpMemoryLimit = memory_limit

    options.hSrcDS = src_ds.this
    options.hSrcDS = dst_ds.this

    if max_error > 0:
        options.pTransformerArg = GDALCreateApproxTransformer(
            GDALGenImgProjTransform, transformer, max_error
        )
        options.pfnTransformer = GDALApproxTransform
    else:
        options.pfnTransformer = GDALGenImgProjTransform
        options.pTransformerArg = transformer

    if options.nBandCount == 0:
        # TODO: implement srcbands
        pass

    # TODO: nodata setup

    warper = GDALCreateWarpOperation(options)
    GDALChunkAndWarpImage(warper, 0, 0, size_x, size_y)

    GDALDestroyWarpOptions(options)


def _open_ds(path_or_ds):
    if isinstance(path_or_ds, string_types):
        gdal.AllRegister()
        return gdal.OpenShared(str(path_or_ds))
    return path_or_ds


def is_extended():
    """ check whether the EOX's GDAL extensions are available
        (True) or not (False)
    """
    return bool(
        GDALCreateTPS2TransformerLSQGrid or GDALCreateTPS2TransformerExt
    )


def suggest_transformer(path_or_ds):
    """ suggest value of method and order to be passed
        to ``get_footprint_wkt`` and ``rect_from_subset``

        :param path_or_ds: a :class:`GDAL Dataset
                           <eoxserver.contrib.gdal.Dataset>` or a path to such.
        :returns: a :class:`dict` containing the keys ``method`` and ``order``
    """

    # get info about the dataset
    ds = _open_ds(path_or_ds)

    nn = ds.GetGCPCount()
    sx = ds.RasterXSize
    sy = ds.RasterYSize

    # guess reasonable limit number of tie-points
    # (Assuming that the tiepoints cover but not execeed
    # the full raster image. That way we don't need
    # to calculate bounding box of the tiepoints' set.)
    nx = 5
    ny = int(max(1, 0.5*nx*float(sy)/float(sx)))
    ng = (nx+1) * (ny+1) + 10

    # check if we deal with an outline along the image's vertical edges
    if nn < 500:  # avoid check for large tie-point sets
        cnt = 0
        for gcp in ds.GetGCPs():
            cnt += (gcp.GCPPixel < 1) or (gcp.GCPPixel >= (sx - 1))
        is_vertical_outline = (cnt == nn)
    else:
        is_vertical_outline = False

    # check whether the GDAL extensions are available
    if is_extended():  # extended GDAL
        # set default to TPS and 3rd order augmenting polynomial
        order = 3
        method = METHOD_TPS

        # some very short ASAR products need 1st order augmenting polynomial
        # the numerics for higher order aug.pol. becomes a bit `wobbly`
        if 4 * sy < sx:
            order = 1

        # small fotprints such as ngEO should use lower TPS-AP order
        if is_vertical_outline:
            order = 1

        # for excessive number of source tiepoints use Least-Square TPS fit
        if nn > ng:
            method = METHOD_TPS_LSQ

    else:  # baseline GDAL

        # set default to TPS and 1st order
        # (the only order available in baseline GDAL)
        order = 1
        method = METHOD_TPS

        # for excessive number of source tiepoints use polynomial GCP fit
        # (the result will most likely incorrect but there is nothing
        # better to be done with the baseline GDAL)
        if nn > ng:
            method = METHOD_GCP
            order = 0  # automatic order selection

    return {'method': method, 'order': order}



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

from django.contrib.gis.geos import Polygon, MultiPolygon

from eoxserver.core.util.xmltools import parse, NameSpace, NameSpaceMap
from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.core.util.iteratortools import pairwise
from eoxserver.core.decoders import xml

NS_GMD = NameSpace("http://www.isotc211.org/2005/gmd", "gmd")
NS_GML = NameSpace("http://www.opengis.net/gml", "gml")
NS_GCO = NameSpace("http://www.isotc211.org/2005/gco", "gco")

nsmap = NameSpaceMap(NS_GMD, NS_GCO, NS_GML)


class InspireFormatReader(object):
    def test(self, obj):
        tree = parse(obj)
        return tree is not None and tree.getroot().tag == NS_GMD("MD_Metadata")

    def read(self, obj):
        tree = parse(obj)
        if tree is not None:
            decoder = InspireFormatDecoder(tree)
            return {
                "identifier": decoder.identifier,
                "begin_time": decoder.begin_time,
                "end_time": decoder.end_time,
                "footprint": decoder.footprint,
                "format": "inspire"
            }
        raise Exception("Could not parse from obj '%s'." % repr(obj))


def parse_line_string(string):
    raw_coords = map(float, string.strip().split())
    return MultiPolygon(
        Polygon([(lon, lat) for lat, lon in pairwise(raw_coords)])
    )


class InspireFormatDecoder(xml.Decoder):
    identifier = xml.Parameter("gmd:fileIdentifier/gco:CharacterString/text()", type=str, num=1)
    begin_time = xml.Parameter("gmd:dataQualityInfo/gmd:DQ_DataQuality/gmd:lineage/gmd:LI_Lineage/gmd:source/gmd:LI_Source/gmd:sourceExtent/gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:beginPosition/text()", type=parse_iso8601, num=1)
    end_time = xml.Parameter("gmd:dataQualityInfo/gmd:DQ_DataQuality/gmd:lineage/gmd:LI_Lineage/gmd:source/gmd:LI_Source/gmd:sourceExtent/gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:endPosition/text()", type=parse_iso8601, num=1)
    footprint = xml.Parameter("gmd:identificationInfo/gmd:MD_DataIdentification/gmd:extent/gmd:EX_Extent/gmd:geographicElement/gmd:EX_BoundingPolygon/gmd:polygon/gml:LineString/gml:posList/text()", type=parse_line_string, num=1)

    namespaces = nsmap



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2014 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

import re

from django.db.models import ForeignKey
from django.contrib.gis.geos import Polygon
from django.contrib.gis.gdal import SpatialReference, CoordTransform
from django.utils.six import string_types

from eoxserver.backends.access import vsi_open
from eoxserver.backends.util import resolve_storage
from eoxserver.resources.coverages import models
from eoxserver.resources.coverages.metadata.coverage_formats import (
    get_reader_by_test
)
from eoxserver.resources.coverages.registration.exceptions import (
    RegistrationError
)


class RegistrationReport(object):
    def __init__(self, coverage, replaced, metadata_parsers,
                 retrieved_metadata):
        self.coverage = coverage
        self.replaced = replaced
        self.metadata_parsers = metadata_parsers
        self.retrieved_metadata = retrieved_metadata


class BaseRegistrator(object):
    """ Abstract base component to be used by specialized registrators.
    """

    abstract = True

    metadata_keys = frozenset((
        "identifier",
        # "footprint", "begin_time", "end_time",
        "size", "origin", "grid"
    ))

    def register(self, data_locations, metadata_locations,
                 coverage_type_name=None, footprint_from_extent=False,
                 overrides=None, identifier_template=None,
                 highest_resolution=False, replace=False, cache=None,
                 use_subdatasets=False, simplify_footprint_tolerance=None):
        """ Main registration method

            :param data_locations:
            :param data_semantics: Either a list of strings (one for each data
                                   location in ``data_locations``) or ``None``,
                                   in which case the semantics will be filled
                                   by best guess.
            :param metadata_locations:
            :param overrides:
            :returns: A registration report
            :rtype: `RegistrationReport`
        """
        replaced = False
        retrieved_metadata = overrides or {}

        # fetch the coverage type if a type name was specified
        coverage_type = None
        if coverage_type_name:
            try:
                coverage_type = models.CoverageType.objects.get(
                    name=coverage_type_name
                )
            except models.CoverageType.DoesNotExist:
                raise RegistrationError(
                    'No such coverage type %r' % coverage_type_name
                )

        # create MetaDataItems for each item that is metadata
        metadata_items = [
            models.MetaDataItem(
                location=location[-1],
                storage=resolve_storage(location[:-1])
            )
            for location in metadata_locations
        ]

        # prepare ArrayDataItems for each given location
        arraydata_items = []
        for location in data_locations:
            # handle storages and/or subdataset specifiers
            path = location[-1]
            # parts = path.split(':')

            parts = [
                part.replace('\\:', ':')
                for part in re.split(r'(?<!\\):', path)
            ]

            subdataset_type = None
            subdataset_locator = None
            if use_subdatasets and len(parts) > 1:
                path = parts[1]
                subdataset_type = parts[0]
                subdataset_locator = ":".join(parts[2:])
            else:
                path = path.replace('\\:', ':')

            arraydata_items.append(
                models.ArrayDataItem(
                    location=path,
                    storage=resolve_storage(location[:-1]),
                    subdataset_type=subdataset_type,
                    subdataset_locator=subdataset_locator,
                )
            )

        metadata_parsers = []

        # read metadata until we are satisfied or run out of metadata items
        for metadata_item in metadata_items:
            if not self.missing_metadata_keys(retrieved_metadata):
                break

            metadata_parsers.append(
                self._read_metadata(
                    metadata_item, retrieved_metadata, cache
                )
            )

        # check the coverage type for expected amount of fields
        if coverage_type:
            num_fields = coverage_type.field_types.count()
            if len(arraydata_items) != 1 \
                    and len(arraydata_items) != num_fields:
                raise RegistrationError(
                    'Invalid number of data files specified. Expected 1 or %d '
                    'got %d.'
                    % (num_fields, len(arraydata_items))
                )

            # TODO: lookup actual band counts

            if len(arraydata_items) == 1:
                arraydata_items[0].band_count = num_fields

            else:
                for i, arraydata_item in enumerate(arraydata_items):
                    arraydata_item.field_index = i
                    arraydata_item.band_count = 1

        elif len(arraydata_items) != 1:
            raise RegistrationError(
                'Invalid number of data files specified.'
            )

            # TODO find actual bands

        # if there is still some metadata missing, read it from the data
        for arraydata_item in arraydata_items:
            if not self.missing_metadata_keys(retrieved_metadata) and \
                    not highest_resolution:
                break
            metadata_parsers.append(
                self._read_metadata_from_data(
                    arraydata_item, retrieved_metadata, cache,
                    highest_resolution
                )
            )

        if self.missing_metadata_keys(retrieved_metadata):
            raise RegistrationError(
                "Missing metadata keys %s."
                % ", ".join(self.missing_metadata_keys(retrieved_metadata))
            )

        identifier = retrieved_metadata["identifier"]
        if identifier_template:
            identifier = identifier_template.format(**retrieved_metadata)
            retrieved_metadata["identifier"] = identifier

        collections = []
        product = None
        if replace:
            try:
                # get a list of all collections the coverage was in.
                coverage = models.Coverage.objects.get(
                    identifier=identifier
                )
                product = coverage.parent_product
                collections = list(models.Collection.objects.filter(
                    coverages=coverage.pk
                ))

                coverage.delete()
                replaced = True

            except models.Coverage.DoesNotExist:
                pass

        # calculate the footprint from the extent
        if footprint_from_extent:
            footprint = self._footprint_from_grid(
                retrieved_metadata['grid'], retrieved_metadata['origin'],
                retrieved_metadata['size']
            )
            retrieved_metadata['footprint'] = footprint

        if simplify_footprint_tolerance is not None and \
                retrieved_metadata.get('footprint'):
            footprint = retrieved_metadata.get('footprint')
            retrieved_metadata['footprint'] = footprint.simplify(
                simplify_footprint_tolerance, preserve_topology=True
            )

        coverage = self._create_coverage(
            identifier=identifier,
            footprint=retrieved_metadata.get('footprint'),
            begin_time=retrieved_metadata.get('begin_time'),
            end_time=retrieved_metadata.get('end_time'),

            size=retrieved_metadata['size'],
            origin=retrieved_metadata['origin'],
            grid=retrieved_metadata['grid'],
            coverage_type_name=coverage_type_name,

            arraydata_items=arraydata_items,
            metadata_items=metadata_items,
        )

        # when we replaced the coverage, re-insert the newly created coverage
        # to the collections and/or product
        for collection in collections:
            models.collection_insert_eo_object(collection, coverage)

        if product:
            models.product_add_coverage(product, coverage)

        return RegistrationReport(
            coverage, replaced, metadata_parsers, retrieved_metadata
        )

    def _read_metadata(self, metadata_item, retrieved_metadata, cache):
        """ Read all available metadata of a ``data_item`` into the
        ``retrieved_metadata`` :class:`dict`.
        """

        with vsi_open(metadata_item) as f:
            content = f.read()
            reader = get_reader_by_test(content)
            if reader:
                values = reader.read(content)

                format_ = values.pop("format", None)
                if format_:
                    metadata_item.format = format_

                for key, value in values.items():
                    retrieved_metadata.setdefault(key, value)

                if values:
                    return reader, values
            return None

    def _read_metadata_from_data(self, data_item, retrieved_metadata, cache,
                                 highest_resolution):
        "Interface method to be overridden in subclasses"
        raise NotImplementedError

    def _footprint_from_grid(self, grid, origin, size):
        "Calculate the footprint from the grid"
        if grid['axis_types'][:2] != ['spatial', 'spatial']:
            raise RegistrationError("Cannot compute footprint from given grid")

        x1, y1 = origin[:2]
        dx, dy = grid['axis_offsets']
        sx, sy = size[:2]
        x2, y2 = (x1 + sx * dx, y1 + sy * dy)

        footprint = Polygon.from_bbox((
            min(x1, x2), min(y1, y2),
            max(x1, x2), max(y1, y2)
        ))

        footprint.transform(
            CoordTransform(
                SpatialReference(grid['coordinate_reference_system']),
                SpatialReference(4326)
            )
        )
        return footprint

    def _create_coverage(self, identifier, footprint, begin_time, end_time,
                         size, origin, grid, coverage_type_name,
                         arraydata_items, metadata_items):

        coverage_type = None
        if coverage_type_name:
            try:
                coverage_type = models.CoverageType.objects.get(
                    name=coverage_type_name
                )
            except models.CoverageType.DoesNotExist:
                raise RegistrationError(
                    'Coverage type %r does not exist' % coverage_type_name
                )

        grid = self._get_grid(grid)

        if len(size) < 4:
            size = list(size) + [None] * (4 - len(size))
        elif len(size) > 4:
            raise RegistrationError('Highest dimension number is 4.')

        if len(origin) < 4:
            origin = list(origin) + [None] * (4 - len(origin))
        elif len(origin) > 4:
            raise RegistrationError('Highest dimension number is 4.')

        (axis_1_size, axis_2_size, axis_3_size, axis_4_size) = size
        (axis_1_origin, axis_2_origin, axis_3_origin, axis_4_origin) = origin

        coverage = models.Coverage(
            identifier=identifier, footprint=footprint,
            begin_time=begin_time, end_time=end_time,
            coverage_type=coverage_type,
            grid=grid,
            axis_1_origin=axis_1_origin,
            axis_2_origin=axis_2_origin,
            axis_3_origin=axis_3_origin,
            axis_4_origin=axis_4_origin,
            axis_1_size=axis_1_size,
            axis_2_size=axis_2_size,
            axis_3_size=axis_3_size,
            axis_4_size=axis_4_size,
        )

        coverage.full_clean()
        coverage.save()

        # attach all data items
        for metadata_item in metadata_items:
            metadata_item.eo_object = coverage
            metadata_item.full_clean()
            metadata_item.save()

        for arraydata_item in arraydata_items:
            arraydata_item.coverage = coverage
            arraydata_item.full_clean()
            arraydata_item.save()

        return coverage

    def _create_metadata(self, coverage, metadata_values):
        metadata_values = dict(
            (name, convert(name, value, models.CoverageMetadata))
            for name, value in metadata_values.items()
            if value is not None
        )

        models.CoverageMetadata.objects.create(
            coverage=coverage, **metadata_values
        )

    def missing_metadata_keys(self, retrieved_metadata):
        """ Return a :class:`frozenset` of metadata keys still missing.
        """
        return self.metadata_keys - frozenset(retrieved_metadata.keys())

    def _get_grid(self, definition):
        return get_grid(definition)


def get_grid(definition):
    """ Get or create a grid according to our defintion
    """
    grid = None
    if isinstance(definition, string_types):
        try:
            grid = models.Grid.objects.get(name=definition)
        except models.Grid.DoesNotExist:
            raise RegistrationError(
                'Grid %r does not exist' % definition
            )
    elif definition:
        axis_names = definition.get('axis_names', [])
        axis_types = definition['axis_types']
        axis_offsets = definition['axis_offsets']

        # check lengths and destructure
        if len(axis_types) != len(axis_offsets):
            raise RegistrationError('Dimensionality mismatch')
        elif axis_names and len(axis_names) != len(axis_types):
            raise RegistrationError('Dimensionality mismatch')

        if len(axis_types) < 4:
            axis_types = list(axis_types) + [None] * (4 - len(axis_types))
        elif len(axis_types) > 4:
            raise RegistrationError('Highest dimension number is 4.')

        if len(axis_offsets) < 4:
            axis_offsets = (
                list(axis_offsets) + [None] * (4 - len(axis_offsets))
            )
        elif len(axis_offsets) > 4:
            raise RegistrationError('Highest dimension number is 4.')

        # translate axis type name to ID
        axis_type_names_to_id = {
            name: id_
            for id_, name in models.Grid.AXIS_TYPES
        }

        axis_types = [
            axis_type_names_to_id[axis_type] if axis_type else None
            for axis_type in axis_types
        ]

        for name, offset in zip(axis_names, axis_offsets):
            if offset == 0:
                raise RegistrationError(
                    'Invalid offset for axis %s: %s.' % (name, offset)
                )

        # unwrap axis types, offsets, names
        (type_1, type_2, type_3, type_4) = axis_types
        (offset_1, offset_2, offset_3, offset_4) = axis_offsets

        # TODO: use names like 'time', or 'x'/'y', etc
        axis_names = axis_names or [
            '%d' % i if i < len(axis_types) else None
            for i in range(len(axis_types))
        ]

        (name_1, name_2, name_3, name_4) = (
            axis_names + [None] * (4 - len(axis_names))
        )

        try:
            # try to find a suitable grid: with the given axis types,
            # offsets and coordinate reference system
            grid = models.Grid.objects.get(
                coordinate_reference_system=definition[
                    'coordinate_reference_system'
                ],
                axis_1_type=type_1,
                axis_2_type=type_2,
                axis_3_type=type_3,
                axis_4_type=type_4,
                axis_1_offset=offset_1,
                axis_2_offset=offset_2,
                axis_3_offset=offset_3,
                axis_4_offset=offset_4,
            )
        except models.Grid.DoesNotExist:
            # create a new grid from the given definition
            grid = models.Grid.objects.create(
                coordinate_reference_system=definition[
                    'coordinate_reference_system'
                ],
                axis_1_name=name_1,
                axis_2_name=name_2,
                axis_3_name=name_3,
                axis_4_name=name_4,
                axis_1_type=type_1,
                axis_2_type=type_2,
                axis_3_type=type_3,
                axis_4_type=type_4,
                axis_1_offset=offset_1,
                axis_2_offset=offset_2,
                axis_3_offset=offset_3,
                axis_4_offset=offset_4,
                resolution=definition.get('resolution')
            )
    return grid


def is_common_value(field):
    try:
        if isinstance(field, ForeignKey):
            field.related_model._meta.get_field('value')
            return True
    except Exception:
        pass
    return False


def convert(name, value, model_class):
    field = model_class._meta.get_field(name)
    if is_common_value(field):
        return field.related_model.objects.get_or_create(
            value=value
        )[0]
    elif field.choices:
        return dict((v, k) for k, v in field.choices)[value]
    return value



#-------------------------------------------------------------------------------
#
# WPS 1.0 execute response XML encoder
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#          Martin Paces <martin.paces@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------
#pylint: disable=too-many-arguments, too-many-locals, bad-continuation

from lxml import etree
from django.utils.timezone import now
from django.utils.six import itervalues

from eoxserver.core.config import get_eoxserver_config
from eoxserver.services.ows.common.config import CapabilitiesConfigReader
from eoxserver.core.util.timetools import isoformat
from eoxserver.services.ows.wps.v10.util import WPS, OWS, ns_xlink, ns_xml

from eoxserver.services.ows.wps.exceptions import OWS10Exception
from eoxserver.services.ows.wps.parameters import (
    Parameter, LiteralData, ComplexData, BoundingBoxData,
    fix_parameter, InputReference, Reference, RequestParameter,
)

from .process_description import encode_process_brief
from .parameters import (
    encode_input_exec, encode_output_exec, encode_output_def
)
from .base import WPS10BaseXMLEncoder
from eoxserver.services.ows.wps.exceptions import InvalidOutputValueError


class WPS10ExecuteResponseXMLEncoder(WPS10BaseXMLEncoder):
    """ WPS 1.0 ExecuteResponse XML response encoder. """

    def __init__(self, process, resp_form, raw_inputs, inputs=None,
                 status_location=None):
        super(WPS10ExecuteResponseXMLEncoder, self).__init__()
        self.process = process
        self.resp_form = resp_form
        self.raw_inputs = raw_inputs
        self.inputs = inputs
        self.status_location = status_location

    def _encode_common(self, status):
        """ Encode common response element. """
        elem = _encode_common_response(
            self.process, status, self.inputs, self.raw_inputs, self.resp_form
        )
        if self.status_location:
            elem.set("statusLocation", self.status_location)
        return elem

    def encode_response(self, results):
        """Encode ProcessSucceeded execute response including the output data."""
        elem = self._encode_common(WPS(
            "ProcessSucceeded",
            "The processes execution completed successfully."
        ))
        outputs = []
        for result, prm, req in itervalues(results):
            outputs.append(_encode_output(result, prm, req))
        elem.append(WPS("ProcessOutputs", *outputs))
        return elem

    def encode_failed(self, exception):
        """ Encode ProcessFailed execute response."""
        # NOTE: Some exceptions such as the urllib2.HTTPError have also
        # the 'code' attribute and the duck typing does not work very well.
        # Therefore we need match the exception base type.
        if isinstance(exception, OWS10Exception):
            code = exception.code
            locator = exception.locator
        else:
            code = "NoApplicableCode"
            locator = type(exception).__name__
        message = str(exception)
        exc_attr = {"exceptionCode": str(code)}
        if locator:
            exc_attr["locator"] = str(locator)
        exc_elem = OWS("Exception", OWS("ExceptionText", message), **exc_attr)
        status = WPS("ProcessFailed", WPS("ExceptionReport", exc_elem))
        return self._encode_common(status)

    def encode_started(self, progress=0, message=None):
        """ Encode ProcessStarted execute response."""
        if not message:
            message = "The processes execution is in progress."
        return self._encode_common(WPS(
            "ProcessStarted", message,
            percentCompleted=("%d" % min(99, max(0, int(float(progress)))))
        ))

    def encode_paused(self, progress=0):
        """ Encode ProcessPaused execute response."""
        return self._encode_common(WPS(
            "ProcessPaused", "The processes execution is paused.",
            percentCompleted=("%d" % min(99, max(0, int(float(progress)))))
        ))

    def encode_accepted(self):
        """ Encode ProcessAccepted execute response."""
        return self._encode_common(WPS(
            "ProcessAccepted", "The processes was accepted for execution."
        ))

#-------------------------------------------------------------------------------

def _encode_common_response(process, status_elem, inputs, raw_inputs, resp_doc):
    """Encode common execute response part shared by all specific responses."""
    inputs = inputs or {}
    conf = CapabilitiesConfigReader(get_eoxserver_config())
    url = conf.http_service_url
    if url[-1] == "?":
        url = url[:-1]
    elem = WPS("ExecuteResponse",
        encode_process_brief(process),
        WPS("Status", status_elem, creationTime=isoformat(now())),
        {
            "service": "WPS",
            "version": "1.0.0",
            ns_xml("lang"): "en-US",
            "serviceInstance": (
                "%s?service=WPS&version=1.0.0&request=GetCapabilities" % url
            )
        },
    )

    if resp_doc.lineage:
        inputs_data = []
        for id_, prm in process.inputs:
            if isinstance(prm, RequestParameter):
                continue
            prm = fix_parameter(id_, prm)
            data = inputs.get(id_)
            rawinp = raw_inputs.get(prm.identifier)
            if rawinp is not None:
                inputs_data.append(_encode_input(data, prm, rawinp))
        elem.append(WPS("DataInputs", *inputs_data))

        outputs_def = []
        for id_, prm in process.outputs:
            prm = fix_parameter(id_, prm)
            outdef = resp_doc.get(prm.identifier)
            if outdef is not None:
                outputs_def.append(encode_output_def(outdef))
        elem.append(WPS("OutputDefinitions", *outputs_def))

    return elem


def _encode_input(data, prm, raw):
    """ Encode one DataInputs sub-element. """
    elem = encode_input_exec(raw)
    if isinstance(raw, InputReference):
        elem.append(_encode_input_reference(raw))
    elif isinstance(prm, LiteralData):
        elem.append(WPS("Data", _encode_raw_input_literal(raw, prm)))
    elif isinstance(prm, BoundingBoxData):
        if data is None:
            data = prm.parse(raw.data)
        elem.append(WPS("Data", _encode_bbox(data, prm)))
    elif isinstance(prm, ComplexData):
        if data is None:
            data = prm.parse(
                data=raw.data, mime_type=raw.mime_type, encoding=raw.encoding,
                schema=raw.schema
            )
        elem.append(WPS("Data", _encode_complex(data, prm)))
    return elem


def _encode_output(data, prm, req):
    """ Encode one ProcessOutputs sub-element. """
    elem = encode_output_exec(Parameter(
        prm.identifier, req.title or prm.title, req.abstract or prm.abstract
    ))
    if isinstance(data, Reference):
        elem.append(_encode_output_reference(data, prm))
    elif isinstance(prm, LiteralData):
        elem.append(WPS("Data", _encode_literal(data, prm, req)))
    elif isinstance(prm, BoundingBoxData):
        elem.append(WPS("Data", _encode_bbox(data, prm)))
    elif isinstance(prm, ComplexData):
        elem.append(WPS("Data", _encode_complex(data, prm)))
    return elem


def _encode_input_reference(ref):
    """ Encode DataInputs/Reference element. """
    #TODO proper input reference encoding
    return WPS("Reference", **{ns_xlink("href"): ref.href})


def _encode_output_reference(ref, prm):
    """ Encode ProcessOutputs/Reference element. """
    #TODO proper output reference encoding
    mime_type = getattr(ref, 'mime_type', None)
    encoding = getattr(ref, 'encoding', None)
    schema = getattr(ref, 'schema', None)
    if mime_type is None and hasattr(prm, 'default_format'):
        default_format = prm.default_format
        mime_type = default_format.mime_type
        encoding = default_format.encoding
        schema = default_format.schema
    attr = {
        #ns_xlink("href"): ref.href,
        'href': ref.href,
    }
    if mime_type:
        attr['mimeType'] = mime_type
    if encoding is not None:
        attr['encoding'] = encoding
    if schema is not None:
        attr['schema'] = schema
    return WPS("Reference", **attr)


def _encode_raw_input_literal(input_raw, prm):
    """ Encode Data/LiteralData element from a raw (unparsed) input ."""
    attrib = {'dataType': prm.dtype.name}
    uom = input_raw.uom or prm.default_uom
    if prm.uoms:
        attrib['uom'] = uom
    return WPS("LiteralData", input_raw.data, **attrib)


def _encode_literal(data, prm, req):
    """ Encode Data/LiteralData element. """
    attrib = {'dataType': prm.dtype.name}
    uom = req.uom or prm.default_uom
    if prm.uoms:
        attrib['uom'] = uom
    try:
        encoded_data = prm.encode(data, uom)
    except (ValueError, TypeError) as exc:
        raise InvalidOutputValueError(prm.identifier, exc)
    return WPS("LiteralData", encoded_data, **attrib)


def _encode_bbox(data, prm):
    """ Encode Data/BoundingBoxData element. """
    try:
        lower, upper, crs = prm.encode_xml(data)
    except (ValueError, TypeError) as exc:
        raise InvalidOutputValueError(prm.identifier, exc)

    return WPS("BoundingBoxData",
        OWS("LowerCorner", lower),
        OWS("UpperCorner", upper),
        crs=crs,
        #dimension="%d"%prm.dimension,
    )
    #NOTE: Although derived from OWS BoundingBox the WPS (schema) does not
    #      allow the dimension attribute.


def _encode_format_attr(data, prm):
    """ Get format attributes of the Data/ComplexData element. """
    mime_type = getattr(data, 'mime_type', None)
    if mime_type is not None:
        encoding = getattr(data, 'encoding', None)
        schema = getattr(data, 'schema', None)
    else:
        deffrm = prm.default_format
        mime_type = deffrm.mime_type
        encoding = deffrm.encoding
        schema = deffrm.schema
    attr = {"mimeType": mime_type}
    if encoding is not None:
        attr['encoding'] = encoding
    if schema is not None:
        attr['schema'] = schema
    return attr


def _encode_complex(data, prm):
    """ Encode Data/ComplexData element. """
    try:
        payload = prm.encode_xml(data)
    except (ValueError, TypeError) as exc:
        raise InvalidOutputValueError(prm.identifier, exc)
    elem = WPS("ComplexData", **_encode_format_attr(data, prm))
    if isinstance(payload, etree._Element):
        elem.append(payload)
    else:
        elem.text = etree.CDATA(payload)
    return elem



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------


import os
import tempfile
import logging
from itertools import chain
import mimetypes

from django.db.models import Q
from django.http import HttpResponse
try:
    from django.http import StreamingHttpResponse
except:
    StreamingHttpResponse = HttpResponse

from django.utils.six import MAXSIZE
from eoxserver.core import Component, implements, ExtensionPoint
from eoxserver.core.config import get_eoxserver_config
from eoxserver.core.decoders import xml, kvp, typelist, enum
from eoxserver.resources.coverages import models
from eoxserver.services.ows.interfaces import (
    ServiceHandlerInterface, GetServiceHandlerInterface,
    PostServiceHandlerInterface
)
from eoxserver.services.ows.wcs.v20.util import (
    nsmap, parse_subset_kvp, parse_subset_xml
)
from eoxserver.services.ows.wcs.v20.parameters import WCS20CoverageRenderParams
from eoxserver.services.ows.common.config import WCSEOConfigReader
from eoxserver.services.ows.wcs.interfaces import (
    WCSCoverageRendererInterface, PackageWriterInterface
)
from eoxserver.services.subset import Subsets, Trim
from eoxserver.services.exceptions import (
    NoSuchDatasetSeriesOrCoverageException, InvalidRequestException,
    InvalidSubsettingException
)


logger = logging.getLogger(__name__)


class WCS20GetEOCoverageSetHandler(Component):
    implements(ServiceHandlerInterface)
    implements(GetServiceHandlerInterface)
    implements(PostServiceHandlerInterface)

    coverage_renderers = ExtensionPoint(WCSCoverageRendererInterface)
    package_writers = ExtensionPoint(PackageWriterInterface)

    service = "WCS"
    versions = ("2.0.0", "2.0.1")
    methods = ['GET', 'POST']
    request = "GetEOCoverageSet"

    index = 21

    def get_decoder(self, request):
        if request.method == "GET":
            return WCS20GetEOCoverageSetKVPDecoder(request.GET)
        elif request.method == "POST":
            return WCS20GetEOCoverageSetXMLDecoder(request.body)

    def get_params(self, coverage, decoder, request):
        return WCS20CoverageRenderParams(
            coverage, Subsets(decoder.subsets), http_request=request
        )

    def get_renderer(self, params):
        for renderer in self.coverage_renderers:
            if renderer.supports(params):
                return renderer

        raise InvalidRequestException(
            "Could not find renderer for coverage '%s'."
        )

    def get_pacakge_writer(self, format, params):
        for writer in self.package_writers:
            if writer.supports(format, params):
                return writer

        raise InvalidRequestException(
            "Format '%s' is not supported." % format, locator="format"
        )

    @property
    def constraints(self):
        reader = WCSEOConfigReader(get_eoxserver_config())
        return {
            "CountDefault": reader.paging_count_default
        }

    def handle(self, request):
        decoder = self.get_decoder(request)
        eo_ids = decoder.eo_ids

        format, format_params = decoder.format
        writer = self.get_pacakge_writer(format, format_params)

        containment = decoder.containment

        count_default = self.constraints["CountDefault"]
        count = decoder.count
        if count_default is not None:
            count = min(count, count_default)

        try:
            subsets = Subsets(
                decoder.subsets,
                crs="http://www.opengis.net/def/crs/EPSG/0/4326",
                allowed_types=Trim
            )
        except ValueError as e:
            raise InvalidSubsettingException(str(e))

        if len(eo_ids) == 0:
            raise

        # fetch a list of all requested EOObjects
        available_ids = models.EOObject.objects.filter(
            identifier__in=eo_ids
        ).values_list("identifier", flat=True)

        # match the requested EOIDs against the available ones. If any are
        # requested, that are not available, raise and exit.
        failed = [eo_id for eo_id in eo_ids if eo_id not in available_ids]
        if failed:
            raise NoSuchDatasetSeriesOrCoverageException(failed)

        collections_qs = subsets.filter(models.Collection.objects.filter(
            identifier__in=eo_ids
        ), containment="overlaps")

        # create a set of all indirectly referenced containers by iterating
        # recursively. The containment is set to "overlaps", to also include
        # collections that might have been excluded with "contains" but would
        # have matching coverages inserted.

        def recursive_lookup(super_collection, collection_set):
            sub_collections = models.Collection.objects.filter(
                collections__in=[super_collection.pk]
            ).exclude(
                pk__in=map(lambda c: c.pk, collection_set)
            )
            sub_collections = subsets.filter(sub_collections, "overlaps")

            # Add all to the set
            collection_set |= set(sub_collections)

            for sub_collection in sub_collections:
                recursive_lookup(sub_collection, collection_set)

        collection_set = set(collections_qs)
        for collection in set(collection_set):
            recursive_lookup(collection, collection_set)

        collection_pks = map(lambda c: c.pk, collection_set)

        # Get all either directly referenced coverages or coverages that are
        # within referenced containers. Full subsetting is applied here.

        coverages_qs = models.Coverage.objects.filter(
            Q(identifier__in=eo_ids) | Q(collections__in=collection_pks)
        )
        coverages_qs = subsets.filter(coverages_qs, containment=containment)

        # save a reference before limits are applied to obtain the full number
        # of matched coverages.
        coverages_no_limit_qs = coverages_qs

        # compute how many (if any) coverages can be retrieved. This depends on
        # the "count" parameter and default setting. Also, if we already
        # exceeded the count, limit the number of dataset series aswell
        """
        if inc_dss_section:
            num_collections = len(collection_set)
        else:
            num_collections = 0

        if num_collections < count and inc_cov_section:
            coverages_qs = coverages_qs.order_by("identifier")[:count - num_collections]
        elif num_collections == count or not inc_cov_section:
            coverages_qs = []
        else:
            coverages_qs = []
            collection_set = sorted(collection_set, key=lambda c: c.identifier)[:count]
        """

        # get a number of coverages that *would* have been included, but are not
        # because of the count parameter
        # count_all_coverages = coverages_no_limit_qs.count()

        # TODO: if containment is "within" we need to check all collections
        # again
        if containment == "within":
            collection_set = filter(lambda c: subsets.matches(c), collection_set)

        coverages = []
        dataset_series = []

        # finally iterate over everything that has been retrieved and get
        # a list of dataset series and coverages to be encoded into the response
        for eo_object in chain(coverages_qs, collection_set):
            if issubclass(eo_object.real_type, models.Coverage):
                coverages.append(eo_object.cast())

        fd, pkg_filename = tempfile.mkstemp()
        tmp = os.fdopen(fd)
        tmp.close()
        package = writer.create_package(pkg_filename, format, format_params)

        for coverage in coverages:
            params = self.get_params(coverage, decoder, request)
            renderer = self.get_renderer(params)
            result_set = renderer.render(params)
            all_filenames = set()
            for result_item in result_set:
                if not result_item.filename:
                    ext = mimetypes.guess_extension(result_item.content_type)
                    filename = coverage.identifier + ext
                else:
                    filename = result_item.filename
                if filename in all_filenames:
                    continue  # TODO: create new filename
                all_filenames.add(filename)
                location = "%s/%s" % (coverage.identifier, filename)
                writer.add_to_package(
                    package, result_item.data_file, result_item.size, location
                )

        mime_type = writer.get_mime_type(package, format, format_params)
        ext = writer.get_file_extension(package, format, format_params)
        writer.cleanup(package)

        response = StreamingHttpResponse(
            tempfile_iterator(pkg_filename), mime_type
        )
        response["Content-Disposition"] = 'inline; filename="ows%s"' % ext
        response["Content-Length"] = str(os.path.getsize(pkg_filename))

        return response


def tempfile_iterator(filename, chunksize=2048, delete=True):
    with open(filename) as file_obj:
        while True:
            data = file_obj.read(chunksize)
            if not data:
                break
            yield data

    if delete:
        os.remove(filename)


def pos_int(value):
    value = int(value)
    if value < 0:
        raise ValueError("Negative values are not allowed.")
    return value


containment_enum = enum(
    ("overlaps", "contains"), False
)


def parse_format(string):
    parts = string.split(";")
    params = dict(
        param.strip().split("=", 1) for param in parts[1:]
    )
    return parts[0], params


class WCS20GetEOCoverageSetKVPDecoder(kvp.Decoder):
    eo_ids      = kvp.Parameter("eoid", type=typelist(str, ","), num=1, locator="eoid")
    subsets     = kvp.Parameter("subset", type=parse_subset_kvp, num="*")
    containment = kvp.Parameter(type=containment_enum, num="?")
    count       = kvp.Parameter(type=pos_int, num="?", default=MAXSIZE)
    format      = kvp.Parameter(num=1, type=parse_format)


class WCS20GetEOCoverageSetXMLDecoder(xml.Decoder):
    eo_ids      = xml.Parameter("/wcseo:EOID/text()", num="+", locator="eoid")
    subsets     = xml.Parameter("/wcs:DimensionTrim", type=parse_subset_xml, num="*")
    containment = xml.Parameter("/wcseo:containment/text()", type=containment_enum, locator="containment")
    count       = xml.Parameter("/@count", type=pos_int, num="?", default=MAXSIZE, locator="count")
    format      = xml.Parameter("/wcs:format/text()", type=parse_format, num=1, locator="format")

    namespaces = nsmap



#-------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
#-------------------------------------------------------------------------------
# Copyright (C) 2013 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#-------------------------------------------------------------------------------

import subprocess
import math
from django.utils.six import string_types

from eoxserver.contrib import gdal, vsi, osr


def get_vrt_driver():
    """ Convenience function to get the VRT driver.
    """
    return gdal.GetDriverByName("VRT")


class VRTBuilder(object):
    """ This class is a helper to easily create VRT datasets from various
    sources.

    :param size_x: the pixel size of the X dimension
    :param size_y: the pixel size of the Y dimension
    :param num_bands: the initial number of bands; bands can be added afterwards
    :param data_type: the GDT data type identifier
    :param vrt_filename: a path the filename shall be stored at; if none is
                         specified the dataset will only be kept in memory
    """

    def __init__(self, size_x, size_y, num_bands=0, data_type=None,
                 vrt_filename=None):
        driver = get_vrt_driver()
        data_type = data_type if data_type is not None else gdal.GDT_Byte
        self._ds = driver.Create(
            vrt_filename or "", size_x, size_y, num_bands, data_type
        )

    @classmethod
    def from_dataset(cls, ds, vrt_filename=None):
        """ A helper function to create a VRT dataset from a given template
        dataset.

        :param ds: a :class:`GDAL Dataset <eoxserver.contrib.gdal.Dataset>`
        """

        vrt_builder = cls(
            ds.RasterXSize, ds.RasterYSize, ds.RasterCount,
            vrt_filename=vrt_filename
        )

        for key, value in ds.GetMetadata().items():
            vrt_builder.dataset.SetMetadataItem(key, value)

    @property
    def dataset(self):
        """ Returns a handle to the underlying VRT :class:`GDAL Dataset
        <eoxserver.contrib.gdal.Dataset>`.
        """
        return self._ds

    def copy_metadata(self, ds):
        """ Copy the metadata fields and values from the given dataset.

        :param ds: a :class:`GDAL Dataset <eoxserver.contrib.gdal.Dataset>`
        """
        for key, value in ds.GetMetadata().items():
            self._ds.SetMetadataItem(key, value)

    def copy_gcps(self, ds, offset=None):
        """ Copy the GCPs from the given :class:`GDAL Dataset
        <eoxserver.contrib.gdal.Dataset>`, optionally offsetting them

        :param ds: a :class:`GDAL Dataset <eoxserver.contrib.gdal.Dataset>`
        :param offset: a 2-tuple of integers; the pixel offset to be applied to
                       any GCP copied
        """
        gcps = ds.GetGCPs()
        if offset:
            gcps = [
                gdal.GCP(
                    gcp.GCPX, gcp.GCPY, gcp.GCPZ,
                    gcp.GCPPixel-offset[0], gcp.GCPLine-offset[1],
                    gcp.Info, gcp.Id
                ) for gcp in gcps
            ]
        self._ds.SetGCPs(gcps, ds.GetGCPProjection())

    def add_band(self, data_type=None, options=None, nodata=None):
        """ Add a band to the VRT Dataset.

        :param data_type: the data type of the band to add. if omitted this is
                          determined automatically by GDAL
        :param options: a list of any string options to be supplied to the new
                        band
        """
        self._ds.AddBand(data_type, options or [])
        if nodata is not None:
            band = self._ds.GetRasterBand(self._ds.RasterCount)
            band.SetNoDataValue(nodata)

    def _add_source_to_band(self, band_index, source):
        band = self._ds.GetRasterBand(band_index)
        if not band:
            raise IndexError

        band.SetMetadataItem("source_0", str(source), "new_vrt_sources")

    def add_simple_source(self, band_index, src, src_band,
                          src_rect=None, dst_rect=None):
        """ Add a new simple source to the VRT.

        :param band_index: the band index the source shall contribute to
        :param src: either a :class:`GDAL Dataset
                    <eoxserver.contrib.gdal.Dataset>` or a file path to the
                    source dataset
        :param src_band: specify which band of the source dataset shall
                         contribute to the target VRT band
        :param src_rect: a 4-tuple of integers in the form (offset-x, offset-y,
                         size-x, size-y) or a :class:`Rect
                         <eoxserver.core.util.rect.Rect>` specifying the source
                         area to contribute
        :param dst_rect: a 4-tuple of integers in the form (offset-x, offset-y,
                         size-x, size-y) or a :class:`Rect
                         <eoxserver.core.util.rect.Rect>` specifying the target
                         area to contribute
        """
        if isinstance(src, string_types):
            pass

        else:
            try:
                src = src.GetFileList()[0]
            except AttributeError:
                raise ValueError("Expected string or GDAL Dataset.")
            except IndexError:
                raise ValueError("Supplied Dataset does not have a filename.")

        lines = [
            "<SimpleSource>",
            '<SourceFilename relativeToVRT="1">%s</SourceFilename>' % src,
            "<SourceBand>%d</SourceBand>" % src_band
        ]
        if src_rect:
            lines.append(
                '<SrcRect xOff="%d" yOff="%d" xSize="%d" ySize="%d"/>'
                % src_rect
            )
        if dst_rect:
            lines.append(
                '<DstRect xOff="%d" yOff="%d" xSize="%d" ySize="%d"/>'
                % dst_rect
            )
        lines.append("</SimpleSource>")

        self._add_source_to_band(band_index, "".join(lines))


class VRTBuilder2(object):
    def __init__(self, size_x, size_y, num_bands=0, data_type=None, vrt_filename=None):
        self.size_x = size_x
        self.size_y = size_y
        self.bands =[]
        self.filename = vrt_filename
        for i in range(0):
            self.add_band(data_type)

        self.transformer = None
        self.geotransform = None

    def set_geotransform(self, geotransform):
        self.geotransform = geotransform

    def add_band(self, data_type=None, options=None, nodata=None):
        self.bands.append(
            dict(data_type=data_type, options=options, nodata=nodata, sources=[])
        )

    def add_simple_source(self, band_index, src, src_band,
                          src_rect=None, dst_rect=None):
        """ Add a new simple source to the VRT.

        :param band_index: the band index the source shall contribute to
        :param src: either a :class:`GDAL Dataset
                    <eoxserver.contrib.gdal.Dataset>` or a file path to the
                    source dataset
        :param src_band: specify which band of the source dataset shall
                         contribute to the target VRT band
        :param src_rect: a 4-tuple of integers in the form (offset-x, offset-y,
                         size-x, size-y) or a :class:`Rect
                         <eoxserver.core.util.rect.Rect>` specifying the source
                         area to contribute
        :param dst_rect: a 4-tuple of integers in the form (offset-x, offset-y,
                         size-x, size-y) or a :class:`Rect
                         <eoxserver.core.util.rect.Rect>` specifying the target
                         area to contribute
        """
        if isinstance(src, string_types):
            pass

        else:
            try:
                src = src.GetFileList()[0]
            except AttributeError:
                raise ValueError("Expected string or GDAL Dataset.")
            except IndexError:
                raise ValueError("Supplied Dataset does not have a filename.")

        self.bands[band_index-1]["sources"].append(
            (src, src_band, src_rect, dst_rect)
        )

    def warped_gcps(self, gcp_dsc, resample="NearestNeighbour", order=0):
        self.transformer = (gcp_ds, resample, order)

    def build_sources(self, sources):
        from lxml.builder import E
        sources = []
        for src, src_band, src_rect, dst_rect in sources:
            elem = E("SimpleSource",
                E("SourceFilename", src, relativeToVRT="0"),
                E("SourceBand", str(src_band))
            )
            if src_rect:
                elem.append(E("SrcRect",
                    xOff=str(src_rect[0]), yOff=str(src_rect[1]),
                    xSize=str(src_rect[2]), ySize=str(src_rect[3]))
                )
            if dst_rect:
                elem.append(E("DstRect",
                    xOff=str(dst_rect[0]), yOff=str(dst_rect[1]),
                    xSize=str(dst_rect[2]), ySize=str(dst_rect[3]))
                )

            sources.append(elem)

        return sources

    def build(self):
        from lxml.builder import E
        from lxml import etree
        root = E("VRTDataset", *[
            E("VRTRasterBand",
                *self.build_sources(band["sources"]),
                dataType=gdal.GetDataTypeName(band["data_type"]),
                band=str(i)
            ) for i, band in enumerate(self.bands, start=1)
        ], rasterXSize=str(self.size_x), rasterYSize=str(self.size_y))
        #root.append()

        if self.transformer:
            gcp_ds, resample, order = self.transformer
            root.append(E("GDALWarpOptions",
                E("WarpMemoryLimit", "6.71089e+07"),
                E("ResampleAlg", resample)
            ))
        return etree.tostring(root, pretty_print=True)


def gdalbuildvrt(filename, paths, separate=False, nodata=None):
    """ Builds a VRT file from the passed files using the
        `gdalbuildvrt` command and stores the resulting file
        under the passed filename.
    """
    args = [
        '/usr/bin/gdalbuildvrt',
        '-resolution',
        'highest',
        '-q',
        '/vsistdout/',
    ]
    if separate:
        args.append('-separate')

    if nodata is not None:
        args.extend([
            '-srcnodata',
            ' '.join(str(value) for value in nodata)
        ])

    content = subprocess.check_output(args + paths)

    with vsi.open(filename, "w") as f:
        f.write(content)


def _determine_parameters(datasets):
    first = datasets[0]
    first_proj = first.GetProjection()
    first_srs = osr.SpatialReference(first_proj)

    first_gt = first.GetGeoTransform()

    others = datasets[1:]

    res_x, res_y = first_gt[1], first_gt[5]
    o_x, o_y = first_gt[0], first_gt[3]

    e_x = o_x + res_x * first.RasterXSize
    e_y = o_y + res_y * first.RasterYSize

    for dataset in others:
        proj = dataset.GetProjection()
        srs = osr.SpatialReference(proj)

        gt = dataset.GetGeoTransform()

        dx, dy = gt[1], gt[5]

        res_x = min(dx, res_x)
        res_y = max(dy, res_y)

        o_x = min(gt[0], o_x)
        o_y = max(gt[3], o_y)

        e_x = max(gt[0] + dx * dataset.RasterXSize, e_x)
        e_y = min(gt[3] + dy * dataset.RasterYSize, e_y)

        assert srs.IsSame(first_srs)
        assert dataset.RasterCount == first.RasterCount

    x_size = int(math.ceil(abs(o_x - e_x) / res_x))
    y_size = int(math.ceil(abs(o_y - e_y) / abs(res_y)))

    return first_proj, (o_x, o_y), (e_x, e_y), (res_x, res_y), (x_size, y_size)


def _get_dst_rect(dataset, o_x, o_y, res_x, res_y):
    gt = dataset.GetGeoTransform()
    dx, dy = gt[1], gt[5]

    x_off = round((gt[0] - o_x) / res_x)
    y_off = round((o_y - gt[3]) / abs(res_y))

    e_x = gt[0] + dx * dataset.RasterXSize
    e_y = gt[3] + dy * dataset.RasterYSize

    x_size = round((e_x - o_x) / res_x) - x_off
    y_size = round((o_y - e_y) / abs(res_y)) - y_off

    return x_off, y_off, x_size, y_size


def mosaic(filenames, save=None):
    """ Creates a mosaic VRT from the specified filenames.
        This function always uses the highest resolution available. The VRT
        is stored under the ``save`` filename, when passed
    """
    datasets = [
        gdal.OpenShared(filename)
        for filename in filenames
    ]

    first = datasets[0]
    proj, (o_x, o_y), _, (res_x, res_y), (size_x, size_y) = \
        _determine_parameters(datasets)

    driver = get_vrt_driver()
    out_ds = driver.Create(save, size_x, size_y, 0)

    out_ds.SetProjection(proj)
    out_ds.SetGeoTransform([o_x, res_x, 0, o_y, 0, res_y])

    for i in range(1, first.RasterCount + 1):
        first_band = first.GetRasterBand(i)
        out_ds.AddBand(first_band.DataType)
        band = out_ds.GetRasterBand(i)
        nodata_value = first_band.GetNoDataValue()
        if nodata_value is not None:
            band.SetNoDataValue(nodata_value)

        for dataset, filename in zip(datasets, filenames):
            x_off, y_off, x_size, y_size = _get_dst_rect(
                dataset, o_x, o_y, res_x, res_y
            )
            nodata_value = dataset.GetRasterBand(i).GetNoDataValue()

            band.SetMetadataItem("source_0", """
                <{source_type}Source>
                    <SourceFilename relativeToVRT="1">{filename}</SourceFilename>
                    <SourceBand>{band}</SourceBand>
                    <SrcRect xOff="0" yOff="0" xSize="{x_size_orig}" ySize="{y_size_orig}"></SrcRect>
                    <DstRect xOff="{x_off}" yOff="{y_off}" xSize="{x_size}" ySize="{y_size}"></DstRect>
                    <NODATA>{nodata_value}</NODATA>
                </{source_type}Source>
            """.format(
                band=i, filename=filename,
                x_size_orig=dataset.RasterXSize,
                y_size_orig=dataset.RasterYSize,
                x_off=x_off, y_off=y_off,
                x_size=x_size, y_size=y_size,
                source_type='Complex' if nodata_value is not None else 'Simple',
                nodata_value=nodata_value if nodata_value is not None else '',
            ), "new_vrt_sources")

    return out_ds


def select_bands(filename, env, band_indices, save=None):
    with gdal.config_env(env):
        ds = gdal.OpenShared(filename)

        out_ds = get_vrt_driver().Create(save, ds.RasterXSize, ds.RasterYSize, 0)
        out_ds.SetProjection(ds.GetProjection())
        out_ds.SetGeoTransform(ds.GetGeoTransform())

        for i, index in enumerate(band_indices, start=1):
            band = ds.GetRasterBand(index)
            out_ds.AddBand(band.DataType)
            out_band = out_ds.GetRasterBand(i)

            nodata_value = band.GetNoDataValue()
            if nodata_value is not None:
                out_band.SetNoDataValue(nodata_value)

            out_band.SetMetadataItem("source_0", """
                <SimpleSource>
                    <SourceFilename relativeToVRT="1">{filename}</SourceFilename>
                    <SourceBand>{band}</SourceBand>
                </SimpleSource>
            """.format(
                band=index, filename=filename
            ), "new_vrt_sources")

    return out_ds


def stack_bands(filenames, env, save=None):
    with gdal.config_env(env):
        datasets = [
            gdal.OpenShared(filename)
            for filename in filenames
        ]

    first = datasets[0]
    proj, (o_x, o_y), _, (res_x, res_y), (size_x, size_y) = \
        _determine_parameters(datasets)

    out_ds = get_vrt_driver().Create(
        save, first.RasterXSize, first.RasterYSize, 0
    )
    out_ds.SetProjection(first.GetProjection())
    out_ds.SetGeoTransform(first.GetGeoTransform())

    out_index = 1
    for dataset, filename in zip(datasets, filenames):
        x_off, y_off, x_size, y_size = _get_dst_rect(
            dataset, o_x, o_y, res_x, res_y
        )

        for index in range(1, dataset.RasterCount + 1):
            band = dataset.GetRasterBand(index)
            out_ds.AddBand(band.DataType)
            out_band = out_ds.GetRasterBand(out_index)

            nodata_value = band.GetNoDataValue()
            if nodata_value is not None:
                out_band.SetNoDataValue(nodata_value)

            out_band.SetMetadataItem("source_0", """
                <SimpleSource>
                    <SourceFilename relativeToVRT="1">{filename}</SourceFilename>
                    <SourceBand>{band}</SourceBand>
                    <SrcRect xOff="0" yOff="0" xSize="{x_size_orig}" ySize="{y_size_orig}"></SrcRect>
                    <DstRect xOff="{x_off}" yOff="{y_off}" xSize="{x_size}" ySize="{y_size}"></DstRect>
                </SimpleSource>
            """.format(
                band=index, filename=filename,
                x_size_orig=dataset.RasterXSize,
                y_size_orig=dataset.RasterYSize,
                x_off=x_off, y_off=y_off,
                x_size=x_size, y_size=y_size,
            ), "new_vrt_sources")

            out_index += 1

    return out_ds

def with_extent(filename, extent, save=None):
    """ Create a VRT and override the underlying files geolocation
    """
    src_ds = gdal.OpenShared(filename)
    width, height = src_ds.RasterXSize, src_ds.RasterYSize
    driver = gdal.GetDriverByName('VRT')
    out_ds = driver.CreateCopy(save, src_ds)

    x = extent[0]
    y = extent[3]

    resx = abs(extent[2] - extent[0]) / width
    resy = abs(extent[3] - extent[1]) / height
    out_ds.SetGeoTransform([
        x,
        resx,
        0,
        y,
        0,
        resy,
    ])
    return out_ds



# ------------------------------------------------------------------------------
#
# Project: EOxServer <http://eoxserver.org>
# Authors: Fabian Schindler <fabian.schindler@eox.at>
#
# ------------------------------------------------------------------------------
# Copyright (C) 2019 EOX IT Services GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies of this Software or works derived from this Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ------------------------------------------------------------------------------

from django.contrib.gis.geos import Polygon, MultiPolygon

from eoxserver.core.util.timetools import parse_iso8601
from eoxserver.core.util.xmltools import NameSpace, NameSpaceMap
from eoxserver.core.util.iteratortools import pairwise
from eoxserver.core.decoders import xml


NS_GSC = NameSpace("http://earth.esa.int/gsc", "gsc")
NS_EOP = NameSpace("http://earth.esa.int/eop", "eop")
NS_OPT = NameSpace("http://earth.esa.int/opt", "opt")
NS_SAR = NameSpace("http://earth.esa.int/sar", "sar")
NS_ATM = NameSpace("http://earth.esa.int/atm", "atm")
NS_GML = NameSpace("http://www.opengis.net/gml", "gml")

nsmap = NameSpaceMap(NS_GML, NS_GSC, NS_EOP, NS_OPT, NS_SAR, NS_ATM)
nsmap_gml = NameSpaceMap(NS_GML)


def parse_ring(string):
    raw_coords = [float(v) for v in string.split()]
    return [(lon, lat) for lat, lon in pairwise(raw_coords)]


def parse_polygon_xml(elem):
    return Polygon(
        parse_ring(
            elem.xpath(
                "gml:exterior/gml:LinearRing/gml:posList", namespaces=nsmap_gml
            )[0].text.strip()
        ), *[
            parse_ring(poslist_elem.text.strip())
            for poslist_elem in elem.xpath(
                "gml:interior/gml:LinearRing/gml:posList", namespaces=nsmap_gml
            )
        ]
    )


class GSCFormatDecoder(xml.Decoder):
    namespaces = nsmap

    identifier = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:metaDataProperty/gsc:EarthObservationMetaData/eop:identifier/text()", type=str, num=1)
    begin_time = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:validTime/gml:TimePeriod/gml:beginPosition/text()", type=parse_iso8601, num=1)
    end_time = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:validTime/gml:TimePeriod/gml:endPosition/text()", type=parse_iso8601, num=1)
    polygons = xml.Parameter("(gsc:sar_metadata|gsc:opt_metadata)/gml:target/eop:Footprint/gml:multiExtentOf/gml:MultiSurface/gml:surfaceMembers/gml:Polygon", type=parse_polygon_xml, num="+")

    @property
    def footprint(self):
        return MultiPolygon(*self.polygons)
